# echidna.tools.train.py

import os, gc
import pickle as pkl
from typing import Callable, Tuple

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import pyro
import pyro.optim as optim
import torch
from pyro.infer import SVI, Trace_ELBO
from tqdm import tqdm

from echidna.tools.utils import (
    EchidnaConfig, 
    EarlyStopping,
    set_sort_order,
)
from echidna.tools.model import Echidna
from echidna.tools.eval import sample
from echidna.tools.housekeeping import save_model, set_posteriors
from echidna.tools.data import (
    build_torch_tensors,
    match_genes,
    train_val_split,
    create_z_pi,
)
from echidna.plot.post import plot_loss
from echidna.utils import get_logger

logger = get_logger(__name__)

def predictive_log_likelihood(echidna: Echidna, data: Tuple):
    guide_trace = pyro.poutine.trace(echidna.guide).get_trace(*data)
    model_trace = pyro.poutine.trace(
        pyro.poutine.replay(echidna.model, trace=guide_trace)
    ).get_trace(*data)
    log_prob = (model_trace.log_prob_sum() - guide_trace.log_prob_sum()).item()
    return log_prob

def echidna_train(adata, Wdf, config=EchidnaConfig()):
    """
    Input
    -----
    adata: sc.Anndata
    Wdf: pd.DataFrame
    config: EchidnaConfig, optional
    
    Output
    ------
    """
    pyro.util.set_rng_seed(config.seed)
    
    num_timepoints = len(adata.obs[config.timepoint_label].unique())
    config.num_timepoints = num_timepoints if config.num_timepoints is None else config.num_timepoints
    config._is_multi = True if config.num_timepoints > 1 else False
    
    config.num_clusters = len(adata.obs[config.clusters].unique())
    
    adata = train_val_split(adata, config)
    
    match_genes(adata, Wdf)
    
    cluster_dtype = adata.obs.loc[:, config.clusters].dtype
    if not pd.api.types.is_integer_dtype(cluster_dtype):
        adata.obs.loc[:, config.clusters + "_categorical"] = adata.obs.loc[:, config.clusters].copy()
        logger.warning(
            f"`{config.clusters}` changed to `{config.clusters}_categorical`."
        )
    adata.obs.loc[:, config.clusters] = pd.Categorical(adata.obs.loc[:,config.clusters]).codes
    
    adata_match = adata[:, adata.var.echidna_matched_genes].copy()
    
    train_data = build_torch_tensors(adata_match[adata_match.obs["echidna_split"]=="train"], config)
    
    if config.val_split > 0:
        val_data = build_torch_tensors(adata_match[adata_match.obs["echidna_split"]=="validation"], config)
    else:
        val_data = None

    config.num_cells = train_data[0].shape[-2]
    config.num_genes = train_data[0].shape[-1]
        
    torch.set_default_dtype(torch.float32)
    torch.set_default_device(config.device)
    
    echidna = train_loop(config, train_data, val_data)
    save_model(adata, echidna, overwrite=True)
    
    del echidna
    gc.collect()
    torch.cuda.empty_cache()
    
def train_loop(config, train_data, val_data):
    pyro.clear_param_store()
    echidna = Echidna(config)
    
    # optimizer = optim.CosineAnnealingLR({
    #     "optimizer": torch.optim.Adam,
    #       "optim_args": {"lr": echidna.config.learning_rate}
    #       , "T_max": 250
    # })
    optimizer = optim.Adam({
        "lr": echidna.config.learning_rate
    })
    
    elbo = Trace_ELBO()
    svi = SVI(echidna.model, echidna.guide, optimizer, loss=elbo)
    
    iterator = tqdm(range(echidna.config.n_steps)) if echidna.config.verbose else range(echidna.config.n_steps)
    best_loss = float("inf")
    
    if (
        echidna.config.patience is not None
        and echidna.config.patience > 0
    ):
        early_stopping = EarlyStopping(patience=echidna.config.patience)
    else:
        early_stopping = EarlyStopping(patience=int(1e30))
        
    patience_counter = 0
    training_loss, validation_loss = [], []
    for _ in iterator:
        try:
            train_elbo = svi.step(*train_data)
            if val_data is not None:
                val_elbo = -predictive_log_likelihood(echidna, val_data)
        except Exception as e:
            logger.error(e)
            echidna = set_posteriors(echidna, train_data)
            return echidna
        if val_data is not None:
            validation_loss.append(val_elbo)
        training_loss.append(train_elbo)
        if echidna.config.verbose:
            avg_loss = np.mean(training_loss[-8:])
            if val_data is not None:
                avg_val_loss = np.mean(validation_loss[-8:])
            else:
                avg_val_loss = 0.
            iterator.set_description(
                f"training loss: {avg_loss:.4f} | "
                f"validation loss: {avg_val_loss:.4f}"
            )
        if early_stopping(avg_val_loss) and val_data is not None:
            break
    if early_stopping.has_stopped() and echidna.config.verbose and val_data is not None:
        logger.info("Early stopping has been triggered.")
    
    echidna = set_posteriors(echidna, train_data)
    
    if echidna.config.verbose:
        plot_loss(training_loss, label="training", log_loss=True)
        if val_data is not None:
            plot_loss(validation_loss, label="validation", log_loss=True)
    
    return echidna

def simulate(adata, overwrite=True):
    """Simulates a training run using sampled X and W.
    
    Parameters
    ----------
    adata : sc.AnnData
        The annotated data matrix with previous training run stored
        in `.uns['echidna']`.
    overwrite : bool
        Whether or not to overwrite previous simulation runs. Default True
        to save memory.
    """
    adata_tmp = adata[adata.obs["echidna_split"] != "discard"].copy()
    config = EchidnaConfig.from_dict(adata_tmp.uns["echidna"]["config"])
    
    X_sample, z_obs = sample(adata_tmp, "X", return_z=True)
    pi_obs = torch.zeros((config.num_timepoints, config.num_clusters))
    for t in range(config.num_timepoints):
        z_tmp = z_obs[t] if len(z_obs.shape) == 2 else z_obs
        counts = torch.bincount(z_obs[t], minlength=config.num_clusters)  # Count occurrences of each cluster
        pi_obs[t] = counts.float() / z_obs.shape[-1]
    z_obs = z_obs.squeeze()
    pi_obs = pi_obs.squeeze()

    W_sample = sample(adata_tmp, "W")
    
    num_cells = z_obs.shape[-1]
    num_val = int(float(config.val_split) * num_cells)
    
    rng = np.random.default_rng(config.seed)
    val_index = rng.choice(range(num_cells), num_val, replace=False)
    train_index = np.setdiff1d(range(num_cells), val_index)
    val_index = torch.from_numpy(val_index).long().to(config.device)
    train_index = torch.from_numpy(train_index).long().to(config.device)
    
    if not config._is_multi:
        z_obs_train = z_obs[train_index]
        z_obs_val = z_obs[val_index]
        X_sample_train = X_sample[train_index, :]
        X_sample_val = X_sample[val_index, :]
    else:
        z_obs_train = z_obs[:, train_index]
        z_obs_val = z_obs[:, val_index]
        X_sample_train = X_sample[:, train_index, :]
        X_sample_val = X_sample[:, val_index, :]
    
    train_data = (X_sample_train, W_sample, pi_obs, z_obs_train)
    val_data = (X_sample_val, W_sample, pi_obs, z_obs_val)

    echidna_sim = train_loop(config, train_data, val_data)
    save_model(adata, echidna_sim, overwrite=overwrite, simulation=True)
    
    del echidna_sim
    torch.cuda.empty_cache()


# echidna.plot.ppc.py

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from scipy.stats import linregress
from scipy.cluster.hierarchy import linkage, leaves_list, dendrogram

import torch
from pyro import render_model

from echidna.tools.housekeeping import load_model, get_learned_params
from echidna.tools.data import build_torch_tensors
from echidna.tools.eval import sample
from echidna.tools.utils import EchidnaConfig, _custom_sort
from echidna.plot.utils import is_notebook, activate_plot_settings
from scipy.stats import linregress, gaussian_kde


def plate_model(adata, filename: str=None):
    """Display plate model
    
    Parameters
    ----------
    adata : sc.AnnData
        The annotated data matrix
    filename : str
        Saves figure to the given path.
    """
    echidna = load_model(adata)
    if echidna.config._is_multi:
        X_rand = torch.randint(low=0, high=40, size=(echidna.config.num_timepoints, 10, echidna.config.num_genes), dtype=torch.float32)
        W_rand = torch.rand((echidna.config.num_timepoints, echidna.config.num_genes), dtype=torch.float32)
        z_rand = torch.randint(low=0, high=echidna.config.num_clusters, size=(echidna.config.num_timepoints, 10,), dtype=torch.int32)
        pi_rand = torch.rand((echidna.config.num_timepoints, echidna.config.num_clusters), dtype=torch.float32)
    else:
        X_rand = torch.randint(low=0, high=40, size=(10, echidna.config.num_genes), dtype=torch.float32)
        W_rand = torch.rand((echidna.config.num_genes), dtype=torch.float32)
        z_rand = torch.randint(low=0, high=echidna.config.num_clusters, size=(10,), dtype=torch.int32)
        pi_rand = torch.rand((echidna.config.num_clusters), dtype=torch.float32)
    
    data=(X_rand, W_rand, pi_rand, z_rand)
    fig = render_model(
        echidna.model, 
        model_args=data, 
        render_params=True, 
        render_distributions=True,
        render_deterministic=True,
        filename=filename,
    )
    if is_notebook():
        display(fig)

def ppc(adata, variable, **kwargs):
    if variable not in ("X", "W", "c", "eta", "cov"):
        raise ValueError(
            "`variable` must be one of or a list of "
            "(\"X\", \"W\", \"c\", \"eta\", \"cov\")"
        )
    activate_plot_settings()
    
    adata_tmp = adata[adata.obs["echidna_split"] != "discard", adata.var.echidna_matched_genes].copy()
    config = EchidnaConfig.from_dict(adata_tmp.uns["echidna"]["config"])
    data = build_torch_tensors(adata_tmp, config)
    learned_params = get_learned_params(load_model(adata_tmp), data)

    ppc_funcs = {
        "X" : ppc_X,
        "W" : ppc_W,
        "c" : ppc_c,
        "eta" : ppc_eta,
        "cov" : ppc_cov,
    }
    
    return ppc_funcs[variable](adata, learned_params, **kwargs)

def ppc_X(adata, learned_params, filename: str=None):
    config = EchidnaConfig.from_dict(adata.uns["echidna"]["config"])
    cell_filter = adata.obs["echidna_split"] != "discard"
    gene_filter = adata.var["echidna_matched_genes"]
    adata_tmp = adata[cell_filter, gene_filter].copy()
    
    # X_true = []
    # if config._is_multi:
    #     timepoints = np.unique(adata_tmp.obs[config.timepoint_label])
    #     if "timepoint_order" in adata_tmp.uns["echidna"]:
    #         timepoints = _custom_sort(
    #             timepoints,
    #             adata_tmp.uns["echidna"]["timepoint_order"]
    #         )
    #     for tp in timepoints:
    #         tp_filter = adata_tmp.obs[config.timepoint_label] == tp
    #         X_true.append(adata_tmp[tp_filter].layers[config.counts_layer])
    #     X_true = np.array(X_true)
    # else:
    #     X_true = adata_tmp.layers[config.counts_layer]

    X_true = []
    for k in learned_params:
        if "X" in k: X_true.append(
            learned_params[k]["value"].detach().cpu().numpy()
        )
    X_true = np.array(X_true).squeeze()
    
    X_learned = sample(adata_tmp, "X").detach().cpu().numpy()
    pred_posterior_check(
        X_learned, X_true, name='X', log_scale=True, R_val=False,
        color_by_density=False, title="Observed vs. reconstructed",
        xlabname="True ", ylabname="Reconstructed ", filename=filename,
    )
    
def ppc_W(adata, learned_params, filename: str=None):
    # config = EchidnaConfig.from_dict(adata.uns["echidna"]["config"])
    # Wdf = adata.var[[c for c in adata.var.columns if "echidna_W_" in c]].dropna()
    # if config._is_multi and "timepoint_order" in adata.uns["echidna"]:
    #     Wdf = _custom_sort(Wdf, adata.uns["echidna"]["timepoint_order"])
    #     W_true = Wdf.values.T
    # else:
    #     W_true = Wdf.values
    
    # In learned_params, W and X were observed.
    W_true = learned_params["W"]["value"].detach().cpu().numpy()
    W_learned = sample(
        adata[:, adata.var.echidna_matched_genes].copy(), "W"
    ).detach().cpu().numpy()
    
    pred_posterior_check(W_learned, W_true, name="W", filename=filename)

def ppc_cov(
    adata,
    learned_params,
    difference: bool=False,
    corr: bool=False,
    filename: str=None,
    cmap: str='plasma_r'
):
    echidna_sim = load_model(adata, simulation=True)
    echidna = load_model(adata)
    
    # Calculate the difference matrix
    cov_matrix_simulated = echidna_sim.cov_posterior
    cov_matrix_real = echidna.cov_posterior
    cov_matrix_diff = cov_matrix_real - cov_matrix_simulated
    
    if corr is True:
        sim_std = torch.sqrt(torch.diag(cov_matrix_simulated))
        real_std = torch.sqrt(torch.diag(cov_matrix_real))
        
        cov_matrix_simulated = cov_matrix_simulated / torch.outer(sim_std, sim_std)
        cov_matrix_real = cov_matrix_real / torch.outer(real_std, real_std)
        
        cov_matrix_simulated[torch.isnan(cov_matrix_simulated)] = 0
        cov_matrix_real[torch.isnan(cov_matrix_real)] = 0
        
        cov_matrix_diff = cov_matrix_real - cov_matrix_simulated
    
    cov_matrix_diff = cov_matrix_diff.detach().cpu().numpy()
    cov_matrix_simulated = cov_matrix_simulated.detach().cpu().numpy()
    cov_matrix_real = cov_matrix_real.detach().cpu().numpy()
    cov_corr_title = "Correlation" if corr is True else "Covariance"
    
    if difference is True:
        # Create heatmaps
        plt.figure(figsize=(18, 6))
        
        # Plot the real covariance matrix
        plt.subplot(1, 3, 1)
        sns.heatmap(cov_matrix_real, cmap=cmap, annot=True, fmt='.2f')
        plt.title(f'{cov_corr_title} Matrix (Real)')
    
        # Plot the simulated covariance matrix
        plt.subplot(1, 3, 2)
        sns.heatmap(cov_matrix_simulated, cmap=cmap, annot=True, fmt='.2f')
        plt.title(f'{cov_corr_title} Matrix (Learned)')
    
        # Plot the difference between the two covariance matrices
        plt.subplot(1, 3, 3)
        sns.heatmap(cov_matrix_diff, cmap=cmap, annot=True, fmt='.2f')
        plt.title('Difference (Real - Learned)')
        plt.tight_layout()
    elif difference is False:
        n = cov_matrix_real.shape[-1]
        df1 = pd.DataFrame(cov_matrix_simulated, columns=[f"Clst {i}" for i in range(n)], index=[f"Clst {i}" for i in range(n)])
        df2 = pd.DataFrame(cov_matrix_real, columns=[f"Clst {i}" for i in range(n)], index=[f"Clst {i}" for i in range(n)])
        
        # Perform hierarchical clustering on the first dataset
        linkage_rows = linkage(df1, method='average', metric='euclidean')
        linkage_cols = linkage(df1.T, method='average', metric='euclidean')
    
        # Get the order of the rows and columns
        row_order = leaves_list(linkage_rows)
        col_order = leaves_list(linkage_cols)
    
        # Reorder both datasets
        df1_ordered = df1.iloc[row_order, col_order]
        df2_ordered = df2.iloc[row_order, col_order]
    
        # Create a grid for the plots
        fig = plt.figure(figsize=(20, 10))
    
        # Define the axes for the first plot
        gs = fig.add_gridspec(3, 4, width_ratios=[0.05, 1, 0.05, 1], height_ratios=[0.2, 1, 0.05], wspace=0.1, hspace=0.1)
        ax_col_dendro1 = fig.add_subplot(gs[0, 1])
        ax_heatmap1 = fig.add_subplot(gs[1, 1])
    
        # Define the axes for the second plot
        ax_col_dendro2 = fig.add_subplot(gs[0, 3])
        ax_heatmap2 = fig.add_subplot(gs[1, 3])
    
        # Plot dendrogram for columns of the first dataset
        dendro_col1 = dendrogram(linkage_cols, ax=ax_col_dendro1, orientation='top', no_labels=True, color_threshold=0)
        ax_col_dendro1.set_xticks([])
        ax_col_dendro1.set_yticks([])
        ax_col_dendro1.set_title(f"Refitted {cov_corr_title}")
    
        # Plot heatmap for the first dataset
        sns.heatmap(df1_ordered, ax=ax_heatmap1, cmap=cmap, cbar=False, xticklabels=False, yticklabels=True)
    
        # Plot dendrogram for columns of the second dataset
        dendro_col2 = dendrogram(linkage_cols, ax=ax_col_dendro2, orientation='top', no_labels=True, color_threshold=0)
        ax_col_dendro2.set_xticks([])
        ax_col_dendro2.set_yticks([])
        ax_col_dendro2.set_title(f"Original {cov_corr_title}")
    
        # Plot heatmap for the second dataset
        sns.heatmap(df2_ordered, ax=ax_heatmap2, cmap=cmap, cbar=False, xticklabels=False, yticklabels=True)
        
    if filename: plt.savefig(filename)
    plt.show()

def ppc_c(adata, learned_params, filename: str=None):
    
    echidna = load_model(adata)
    
    c_learned = learned_params["c"]["value"].flatten().detach().cpu().numpy()
    c_posterior = echidna.c_posterior.flatten().detach().cpu().numpy()
    c_learned, c_posterior = _sample_arrays(c_learned, c_posterior, seed=echidna.config.seed)
    
    del echidna
    
    slope, intercept, r_value, p_value, std_err = linregress(c_learned, c_posterior)
    r_squared = r_value

    data = pd.DataFrame({'c_learned': c_learned, 'c_posterior': c_posterior})

    pred_posterior_check(c_learned, c_posterior, "c", equal_line=False, title="Refitted vs. posterior ground truth ",
                         xlabname="Refitted ", ylabname="Posterior ground truth ", filename=filename)

    #plt.figure(figsize=(10, 6))
    #regplot = sns.regplot(data=data, x='c_learned', y='c_posterior', scatter_kws={'s': 10, 'color': 'blue'}, line_kws={'color': 'red'})
    #scatter = sns.scatterplot(data=data, x='c_learned', y='c_posterior', s=50, color='blue', alpha=0.6)

    #plt.text(0.05, 0.95, f'$Pearson R = {r_squared:.4f}$', transform=plt.gca().transAxes,
             #fontsize=12, verticalalignment='top')

    #plt.title('c fitted vs. c posterior truth')
    #plt.xlabel('c fitted')
    #plt.ylabel('c posterior truth')
    
    #if filename: plt.savefig(filename)
    #plt.show()


def ppc_eta(adata, learned_params, filename: str=None):
    echidna_sim = load_model(adata, simulation=True)
    echidna = load_model(adata)
    
    eta_learned = echidna_sim.eta_posterior.flatten().detach().cpu().numpy()
    eta_posterior = echidna.eta_posterior.flatten().detach().cpu().numpy()
    
    eta_learned, eta_posterior = _sample_arrays(eta_learned, eta_posterior, seed=echidna.config.seed)
    
    #slope, intercept, r_value, p_value, std_err = linregress(eta_learned, eta_posterior)
    #r_squared = r_value

    data = pd.DataFrame({'eta_learned': eta_learned, 'eta_posterior': eta_posterior})

    pred_posterior_check(eta_learned, eta_posterior, "$\eta$", equal_line=False, title="Refitted vs. posterior ground truth ",
                         xlabname="Refitted ", ylabname="Posterior ground truth ", filename=filename)

    #scatter = sns.scatterplot(data=data, x='eta_learned', y='eta_posterior', s=50, color='blue', alpha=0.6)
    #contour = sns.kdeplot(data=data, x='eta_learned', y='eta_posterior', levels=10, color='red', linewidths=1.5)

    #for child in scatter.get_children():
     #if isinstance(child, plt.Line2D): 
        #child.set_rasterized(True)

    #plt.text(
        #0.05, 0.95,
        #f'$Pearson R = {r_squared:.4f}$',
        #transform=scatter.transAxes,
        #fontsize=12,
        #verticalalignment='top',
    #)

    #scatter.set_title('eta fitted vs. eta posterior truth')
    #scatter.set_xlabel('eta fitted')
    #scatter.set_ylabel('eta posterior truth')

    #if filename: plt.savefig(filename)
    #plt.show()
    
    #del echidna
    #del echidna_sim

def _sample_arrays(learned, observed, max_samples=int(3e4), seed=42):
    rng = np.random.default_rng(seed)
    total_samples = min(len(learned), max_samples)
    indices = np.random.choice(len(learned), total_samples, replace=False)
    
    learned = learned[indices]
    observed = observed[indices]
    
    return learned, observed    

def pred_posterior_check(
        X_learned: np.ndarray,
        X_true: np.ndarray,
        name: str = "",
        log_scale: bool = False,
        R_val: bool = True,
        equal_line: bool = True,
        color_by_density: bool = False,
        title: str = "Predictive Posterior Check",
        xlabname: str = "True ",
        ylabname: str = "Simulated ",
        filename: str = None,
        subsample: bool = True
):
    if subsample:
      # Subsample for plotting
      num = min(len(X_learned.flatten()), 200000)
      indx = np.random.choice(
          np.arange(len(X_learned.flatten())), num, replace=False)
      X_learned = X_learned.flatten()[indx]
      X_true = X_true.flatten()[indx]

    if log_scale:
        X_learned = np.log(X_learned + 1)
        X_true = np.log(X_true + 1)
        lbl_pstfix = "[log(x + 1)] "
    else:
        lbl_pstfix = ""

    x = X_true
    y = X_learned

    # Perform linear regression
    if R_val:
        slope, intercept, r_value, p_value, std_err = linregress(x, y)
        r2 = r_value**2
        y_pred = slope * x + intercept

    maximum = max(np.max(x), np.max(y))
    minimum = min(np.min(x), np.min(y))

    # Scatter plot
    if color_by_density:
        # Calculate point densities
        xy = np.vstack([x, y])
        z = gaussian_kde(xy)(xy)
        # Plot using density as color
        plt.scatter(x, y, c=z, cmap='viridis',
                    alpha=0.5, label='Data points', s=10, rasterized=True)
    else:
        plt.scatter(x, y, alpha=0.6, c='blue',  rasterized=True)

    if equal_line:
        plt.plot([minimum, maximum], [minimum, maximum], "r", label="x=y")

    # Plot the regression line
    #if R_val:
        #plt.plot(x, y_pred, label="Regression line", color='blue')

    plt.xlabel(xlabname + lbl_pstfix + name, fontsize=16)
    plt.ylabel(ylabname + lbl_pstfix + name, fontsize=16)

    # Annotate the plot with the R^2 value
    if R_val:
        plt.text(0.05, 0.95, f'$PearsonR$ = {r2:.3f}', ha='left', va='center', transform=plt.gca(
        ).transAxes, fontsize=20, bbox=dict(facecolor='white', alpha=0.5))

    plt.legend(loc="lower right")
    plt.title(title + " " + name, fontsize = 16)

    if filename is not None:
        plt.savefig(filename, format='svg')

    plt.show()



# echidna.tools.infer_gd.py

import logging, os, re, sys

import scanpy as sc
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import torch
import pyro
from typing import List, Any, Dict, Optional

from scipy.stats import ttest_ind, mode
from scipy.ndimage import gaussian_filter1d
from hmmlearn import hmm
from sklearn.mixture import GaussianMixture

from echidna.tools.housekeeping import load_model
from echidna.tools.data import filter_low_var_genes, sort_chromosomes
from echidna.tools.eval import sample
from echidna.utils import get_logger, ECHIDNA_GLOBALS

logger = get_logger(__name__)

def cnv_results(adata: sc.AnnData) -> pd.DataFrame:
    echi_cnv_save_path = None
    if "echi_cnv" not in adata.uns["echidna"]["save_data"]:
        raise ValueError(
            "Must run `ec.tl.echi_cnv` first."
        )
    echi_cnv_save_path = adata.uns["echidna"]["save_data"]["echi_cnv"]
    if not os.path.exists(echi_cnv_save_path):
        raise ValueError(
            "Saved results not found. Run `ec.tl.echi_cnv` first."
        )
    echi_cnv = pd.read_csv(echi_cnv_save_path, index_col="geneName")
    
    neutral_save_path = adata.uns["echidna"]["save_data"]["gmm_neutrals"]
    neutral_states = pd.read_csv(
        neutral_save_path,
        index_col="eta_column_label",
    )
    num_clusters = adata.obs[adata.uns["echidna"]["config"]["clusters"]].nunique()
    clone_cols = [f"echidna_clone_{x}" for x in range(num_clusters)]
    
    echidna = load_model(adata)
    eta = echidna.eta_posterior
    del echidna
    
    echidna_matched_genes = adata[:, adata.var["echidna_matched_genes"]].var.index
    eta = eta.T.detach().cpu().numpy()
    eta = pd.DataFrame(eta).set_index(echidna_matched_genes)
    clone_cols = [f"echidna_clone_{c}" for c in range(eta.shape[1])]
    eta.columns = clone_cols

    common_genes = echi_cnv.index.intersection(eta.index)
    eta = eta.loc[common_genes]
    
    for i, c in enumerate(clone_cols):
        echi_cnv[c] = eta[c]
        echi_cnv[c] -= neutral_states.loc[c, "neutral_value_mean"].item()
    
    return echi_cnv

def echi_cnv(
    adata: sc.AnnData,
    genome: pd.DataFrame=None,
    gaussian_smoothing: bool=True,
    n_hmm_components: int=5,
    n_gmm_components: int=5,
    filter_genes: bool=True,
    filter_quantile: float=.7,
    smoother_sigma: float=6,
    smoother_radius: float=8,
    **kwargs,
) -> None:
    if genome is None:
        try:
            genome = pd.read_csv(
                "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/wgEncodeGencodeCompV46.txt.gz",
                delimiter="\t",
                header=None,
                names=["gene_id", "transcript_id", "chrom", "strand", "txStart", "txEnd", "cdsStart", "cdsEnd", "exonCount", "exonStarts", "exonEnds", "score", "geneName", "cdsStartStat", "cdsEndStat", "exonFrames"],
            )
            genome["chrom"] = "chr" + genome.chrom.str.extract(r"chr(\d+|X|Y)")
            genome = genome.groupby(["chrom", "geneName"]).agg(
                txStart=("txStart", "min"),
                txEnd=("txEnd", "max")
            ).reset_index()
            logger.info(
                "`genome` not set, defaulting to hg38"
                "wgEncodeGencodeCompV46."
            )
        except Exception as e:
            logger.info(
                "`genome` not set, defaulting to hg38"
                "wgEncodeGencodeCompV46."
            )
            logger.error(e)
            raise IOError("Must enable internet connection to fetch default genome data.")
    
    if "band" in genome.columns and "chr" not in genome["band"].iloc[0]:
        genome["band"] = genome["chrom"] + "_" + genome["band"]
    genome = sort_chromosomes(genome)
    
    echidna = load_model(adata)
    random_state = echidna.config.seed
    eta = echidna.eta_posterior
    
    del echidna
    torch.cuda.empty_cache()
    
    echidna_matched_genes = adata[:, adata.var["echidna_matched_genes"]].var.index
    
    eta = eta.T.detach().cpu().numpy()
    eta = pd.DataFrame(eta).set_index(echidna_matched_genes)
    clone_cols = [f"echidna_clone_{c}" for c in range(eta.shape[1])]
    eta.columns = clone_cols
    
    adata_filtered = filter_low_var_genes(adata.copy(), quantile=filter_quantile)
    common_genes = adata_filtered.var.index.intersection(
        eta.index
    ).intersection(genome.loc[:, "geneName"])
    genome_unique = genome.drop_duplicates(subset="geneName")
    eta_filtered = eta.reindex(
        genome_unique[genome_unique.loc[:, "geneName"].isin(common_genes)].loc[:, "geneName"]
    )
    
    # GMM gets eta filtered for genes and gaussian smoothed
    eta_filtered_smooth = gaussian_filter1d(
        eta_filtered, sigma=smoother_sigma, axis=0, radius=smoother_radius
    )
    neutral_states = _get_neutral_state(
        eta_filtered_smooth,
        clone_cols,
        n_components=n_gmm_components,
        random_state=random_state,
        **kwargs
    )
    
    # User has choice for HMM on filtering genes and gaussian smoothing
    if filter_genes:
        eta = eta_filtered
    if gaussian_smoothing:
        eta.values[:] = gaussian_filter1d(
            eta, sigma=smoother_sigma, axis=0, radius=smoother_radius
        )

    # OLD - normalize out variance of each gene from eta
    # eta = eta / torch.sqrt(torch.diag(torch.cov(eta.T)))

    eta_genome_merge = genome[["geneName", "chrom"]].merge(
        eta, left_on="geneName",
        right_index=True,
        how="inner",
        validate="many_to_one",
    )
    
    ## OLD
    # if "weight" in genome.columns:
    #     bands_eta_merge[clone_cols] = bands_eta_merge[clone_cols].mul(
    #         bands_eta_merge["weight"], axis="index"
    #     )
    
    # bands_eta_means = bands_eta_merge.groupby("band")[clone_cols].mean()
    # bands_eta_means = bands_eta_means.reindex(
        # genome["band"].drop_duplicates()
    # ).dropna()
    
    for c in clone_cols:
        eta_genome_merge[f"states_{c}"] = _get_states(
            eta_genome_merge.loc[:, c].values,
            n_components=n_hmm_components,
            neutral_mean=neutral_states.loc[c, "neutral_value_mean"].item(),
            neutral_std=neutral_states.loc[c, "neutral_value_std"].item(),
            **kwargs
        )

    echi_cnv_save_path = os.path.join(
        ECHIDNA_GLOBALS["save_folder"],
        adata.uns["echidna"]["run_id"],
        "_echidna_cnv.csv",
    )
    neutrals_save_path = os.path.join(
        ECHIDNA_GLOBALS["save_folder"],
        adata.uns["echidna"]["run_id"],
        "_gmm_neutrals.csv",
    )
    
    eta_genome_merge.to_csv(
        echi_cnv_save_path,
        index=False,
    )
    neutral_states.to_csv(
        neutrals_save_path,
        index=True,
    )
    adata.uns["echidna"]["save_data"]["echi_cnv"] = echi_cnv_save_path
    adata.uns["echidna"]["save_data"]["gmm_neutrals"] = neutrals_save_path
    logger.info(
        "Added `.uns['echidna']['save_data']['echi_cnv']` : Path to CNV inference results.\n"
        "Added `.uns['echidna']['save_data']['gmm_neutrals']` : Path to Echidna cluster neutral value results."
    )

def _get_states(
    vals: np.ndarray,
    n_components: int=5,
    neutral_mean: float=2.0,
    neutral_std: float=1.0,
    transmat_prior: Optional[np.ndarray]=1,
    startprob_prior: Optional[np.ndarray]=1,
    n_iter: int=100,
    verbose: bool=False,
    **args: Dict[str, Any],
) -> List[str]:
    """Implements a Gaussian HMM to call copy number states smoothing along the genome
    
    Parameters
    ----------
    vals: ordered copy number values (from bin_by_bands function)
    n_components: number of components to use for the hmm. We default to 5 for better sensitivity.
    neutral: the number to use for neutral. 
    transmat_prior: Parameters of the Dirichlet prior distribution for each row of the transition probabilities
    startprob_prior: Parameters of the Dirichlet prior distribution for startprob_
    
    Returns
    -------
    pd.DataFrame
    """
    scores = []
    models = []

    idx = np.random.choice(range(100), size=10)
    for i in range(10):
        model = hmm.GaussianHMM(
            n_components=n_components,
            random_state=idx[i],
            n_iter=n_iter,
            transmat_prior=transmat_prior,
            startprob_prior=startprob_prior,
        )
        model.fit(vals.reshape(-1, 1))
        models.append(model)
        scores.append(model.score(vals.reshape(-1, 1)))
        if verbose:
            logger.info(
                f"Converged: {model.monitor_.converged}\t\tScore: {scores[-1]}"
            )
    
    # Get the best model
    model = models[np.argmax(scores)]
    if verbose:
        logger.info(
            f"The best model had a score of {max(scores)}"
            "and {model.n_components} components"
        )
    
    states = model.predict(vals[:, None])

    tmp = pd.DataFrame({"vals": vals, "states": states})
    state_dict = tmp.groupby("states")["vals"].mean().reset_index()

    # state_dict["sq_dist_to_neut"] = (state_dict["vals"] - neutral)**2
    # state_dict["dist_to_neut"] = state_dict["vals"] - neutral
    # state_dict = state_dict.sort_values(by="sq_dist_to_neut")
    
    n_stddevs = 1.96
    state_dict["neutral"] = abs(state_dict["vals"] - neutral_mean) <= n_stddevs * neutral_std
    state_dict["amp"] = state_dict["vals"] - neutral_mean > n_stddevs * neutral_std
    state_dict["del"] = state_dict["vals"] - neutral_mean < -n_stddevs * neutral_std

    # Check if any other states are neutral using a t-test
    # neutral_state = [state_dict.iloc[0]["states"]]
    # neutral_val = [state_dict.iloc[0]["vals"]]
    # neutral_dist = tmp[tmp["states"] == neutral_state[0]]["vals"]

    # for i, row in state_dict.iterrows():
    #     if row["states"] not in neutral_state:
    #         tmp_dist = tmp[tmp["states"] == row["states"]]["vals"]
    #         p_val = ttest_ind(neutral_dist, tmp_dist).pvalue
    #         if p_val > p_thresh:
    #             neutral_state.append(row["states"])
    #             neutral_val.append(row["vals"])
    
    # amp if CN > netural, del if less than
    def classify_state(row):
        if row["neutral"] is True:
            return "neut"
        elif row["amp"] is True:
            return "amp"
        elif row["del"] is True:
            return "del"
    state_dict["CNV"] = state_dict.apply(classify_state, axis=1)
    
    # Map the CNV states back to the original states
    state_map = state_dict.set_index("states")["CNV"].to_dict()
    cnvs = list(map(state_map.get, states))

    return cnvs

def _get_neutral_state(
    eta_filtered_smooth: np.ndarray,
    eta_column_labels: List[str],
    n_components: int=5,
    plot_gmm: bool=False,
    random_state: int=None,
    **args,
) -> pd.DataFrame:
    gmm_means_df = pd.DataFrame(columns=["eta_column_label", "mode", "neutral_value_mean", "neutral_value_std"])
    
    for i, col in enumerate(eta_column_labels):
        cur_vals_filtered = eta_filtered_smooth[:, i].reshape(-1,1)
        gmm = GaussianMixture(n_components=n_components, random_state=random_state).fit(cur_vals_filtered)
        labels = gmm.predict(cur_vals_filtered)
        neut_component = mode(labels, keepdims=False).mode

        gmm_mean = np.mean(cur_vals_filtered[labels == neut_component])
        gmm_std = np.std(cur_vals_filtered[labels == neut_component])

        if plot_gmm:
            _plot_gmm_clusters(
                gmm, cur_vals_filtered, gmm_mean, i
            )

        gmm_means_df.loc[i, :] = [col, neut_component, gmm_mean, gmm_std]

    return gmm_means_df.set_index("eta_column_label")

def _plot_gmm_clusters(gmm, vals_filtered, gmm_mean, i):
    fig, ax = plt.subplots(
            nrows=1,
            ncols=1, 
            figsize=(10, 5)
    )

    x = np.linspace(-5, 10, 1000)
    data = np.asarray(vals_filtered).squeeze()

    sns.histplot(data, bins=30, stat='density', alpha=0.6, color='gray', ax=ax)

    for mean, variance, weight in zip(gmm.means_, gmm.covariances_, gmm.weights_):
        variance = np.sqrt(variance).flatten()
        label = f'Component mean={mean[0]:.3f}'
        color = 'red' if np.isclose(mean[0], gmm_mean, atol=0.1) else None
        linewidth = 3 if color == 'red' else 1

        ax.plot(
            x,
            weight * (1 / np.sqrt(2 * np.pi * variance)) * np.exp(-(x - mean) ** 2 / (2 * variance)),
            label=f'Neutral mean = {gmm_mean:.3f}' if color else label,
            color=color,
            linewidth=linewidth
        )

    logprob = gmm.score_samples(x.reshape(-1, 1))
    pdf = np.exp(logprob)
    ax.plot(x, pdf, '-k', label='Global Density')

    ax.set_xlabel('Eta values')
    ax.set_ylabel('Density')
    ax.set_title(f"Echidna Cluster {i}")
    ax.legend()

    plt.tight_layout()
    plt.show()

def gene_dosage_effect(
    adata,
    **kwargs
):
    """
    eta_mode [n_clusters, 1]
    eta_samples, eta_mean [n_genes, n_clusters]
    c_shape  [n_genes, n_timepoints]
    """
    
    # Retrieve save data from tools functions
    if "gmm_neutrals" not in adata.uns["echidna"]["save_data"]:
        raise ValueError("Must run `ec.tl.echi_cnv` first.")
    neutral_save_path = adata.uns["echidna"]["save_data"]["gmm_neutrals"]
    if not os.path.exists(neutral_save_path):
        raise ValueError(
            "Saved results not found. Run `ec.tl.echi_cnv` first."
        )
    neutral_states = pd.read_csv(neutral_save_path)
    
    model = load_model(adata)

    # echidna_matched_genes = adata_filtered[
    #     :, adata_filtered.var["echidna_matched_genes"]
    # ].var.index
    
    # eta = pd.DataFrame(
    #     model.eta_posterior.T.cpu().detach().numpy(),
    #     index=echidna_matched_genes,
    # )
    # echidna_matched_genes = echidna_matched_genes.intersection(adata_filtered.var.index)
    
    # eta = eta.reindex(echidna_matched_genes)
    # clone_cols = [f"echidna_clone_{c}" for c in range(eta.shape[1])]

    # eta_filtered_smooth = gaussian_filter1d(
    #     eta, sigma=smoother_sigma, axis=0, radius=smoother_radius
    # )
    # eta_mode = _get_neutral_state(
    #     eta_filtered_smooth, clone_cols, n_gmm_components
    # )["neutral_value_mean"]
    
    eta_samples = sample(adata, "eta")
    c_shape = pyro.param("c_shape")
    eta_mean = model.eta_posterior.T #pyro.param("eta_mean")
    eta_mode = torch.tensor(
        neutral_states["neutral_value_mean"].values,
        device=model.config.device
    )
    
    if adata.uns["echidna"]["config"]["_is_multi"]:
        c_shape = c_shape.squeeze() # [T, G]
    c_shape = c_shape.T             # [G, T]
    
    c_shape_expanded = c_shape[None, :, :, None]        # [1, G, T, 1]
    eta_samples_expanded = eta_samples[:, :, None, :]   # [n_samples, G, 1, C]

    # delta Var(c|eta)
    eta_mean_expanded = eta_mean[:, None, :]            # [G, 1, C]
    eta_mode_expanded = eta_mode[None, None, :]         # [1, 1, C]

    delta_var = (
        c_shape[:, :, None] * eta_mean_expanded**2
        - c_shape[:, :, None] * eta_mode_expanded**2
    )  # [G, T, C]

    # Var(E[c|eta])
    exp_c_given_eta = c_shape_expanded * eta_samples_expanded  # [n_samples, G, T, C]
    var_exp_c_given_eta = torch.var(exp_c_given_eta, unbiased=True, dim=0)  # [G, T, C]

    # E[Var(c|eta)]
    exp_var_c_given_eta = c_shape_expanded * eta_samples_expanded**2        # [n_samples, G, T, C]
    exp_var_c_given_eta = torch.mean(exp_var_c_given_eta, dim=0)            # [G, T, C]

    # (Var(c|eta)-Var(c|eta_mode))/(Var(E[c|eta])+E[Var(c|eta)])
    var_exp = delta_var / (var_exp_c_given_eta + exp_var_c_given_eta)       # [G, T, C]
    
    gene_dosage_cache_path = os.path.join(
        ECHIDNA_GLOBALS["save_folder"],
        adata.uns["echidna"]["run_id"],
        "_gene_dosage_cache.pt",
    )
    torch.save(var_exp, f=gene_dosage_cache_path)
    adata.uns["echidna"]["save_data"]["gene_dosage"] = gene_dosage_cache_path
    logger.info(
        "Added `.uns['echidna']['save_data']['gene_dosage']`"
        " : Path to gene dosage effect results."
    )
    
    # return var_exp
    
def genes_to_bands(genes, cytobands):
    spanning_genes = []
    for i, gene in genes.iterrows():
        chrom = gene["chrom"]
        gene_start = gene["txStart"]
        gene_end = gene["txEnd"]
        
        lh_overlap = (
            (gene_start <=  cytobands["bandStart"]) &
            (gene_end >= cytobands["bandStart"])
        )
        in_between = (gene_start >= cytobands["bandStart"]) & (gene_end <= cytobands["bandEnd"])
        rh_overlap = (gene_start <= cytobands["bandEnd"]) & (gene_end >= cytobands["bandEnd"])
        
        bands = cytobands[(chrom == cytobands["chrom"]) & (lh_overlap | in_between | rh_overlap)].copy()
        if len(bands) > 0:
            span = gene_end - gene_start
            bands.loc[:, "weight"] = 1.
            bands.loc[:, "weight"] = np.where(rh_overlap[bands.index], (bands["bandEnd"] - gene_start) / span, bands["weight"])
            bands.loc[:, "weight"] = np.where(lh_overlap[bands.index], (gene_end - bands["bandStart"]) / span, bands["weight"])
            for j, band in bands.iterrows():
                spanning_genes.append((
                    band["chrom"],
                    band["band"],
                    band["bandStart"],
                    band["bandEnd"],
                    gene["geneName"],
                    gene_start,
                    gene_end,
                    band["weight"],
                ))
        del bands
    genome = pd.DataFrame(spanning_genes, columns=["chrom", "band", "bandStart", "bandEnd", "geneName", "txStart", "txEnd", "weight"])
    
    return genome.drop_duplicates(["chrom", "band", "geneName"])



# echidna.tools.__init__.py

from .utils import EchidnaConfig, reset_echidna_memory, set_sort_order
from .data import pre_process, build_torch_tensors, filter_low_var_genes
from .train import echidna_train, simulate
from .eval import echidna_clones, sample, echidna_status
from .infer_gd import genes_to_bands, echi_cnv, cnv_results, gene_dosage_effect
from .housekeeping import save_model, load_model, get_learned_params

__all__ = [
    "EchidnaConfig",
    "reset_echidna_memory",
    "set_sort_order",
    "pre_process",
    "build_torch_tensors",
    "filter_low_var_genes",
    "echidna_train",
    "simulate",
    "echidna_clones",
    "echidna_status",
    "sample",
    "genes_to_bands",
    "echi_cnv",
    "cnv_results",
    "gene_dosage_effect",
    "save_model",
    "load_model",
    "get_learned_params",
]


# echidna.tools.custom_dist.py

import torch
from pyro import distributions as dist
from torch.distributions import constraints

class TruncatedNormal(dist.Distribution):

    '''
    Truncated Normal distribution implemented for Pyro. Derived from numpyro's implementation.
    '''

    def __init__(self, loc, scale, lower=None, upper=None, batch_shape=()):
        super().__init__()
        self.lower = lower if lower is not None else -torch.inf
        self.upper = upper if upper is not None else torch.inf

        loc = loc if isinstance(loc, torch.Tensor) else torch.tensor(loc)
        scale = scale if isinstance(scale, torch.Tensor) else torch.tensor(scale)

        self.batch_shape = torch.broadcast_shapes(
            loc.shape, scale.shape,
        )

        self.base_dist = dist.Normal(loc, scale).expand(self.batch_shape)
        self.event_shape = self.base_dist.event_shape
        self.event_dim = self.base_dist.event_dim

        if self.lower == -torch.inf and self.upper != torch.inf:
            self.support = dist.constraints.less_than(upper)
        elif self.lower != -torch.inf and self.upper == torch.inf:
            self.support = dist.constraints.greater_than(lower)
        else:
            self.support = dist.constraints.interval(lower, upper)

    def expand(self, batch_shape: torch.Size, _instance=None):
        return self.__class__(self.base_dist.loc.expand(batch_shape), self.base_dist.scale.expand(batch_shape), self.lower, self.upper)


    def log_prob(self, value):
        if self.lower == -torch.inf and self.upper != torch.inf:
            # right bound -- (-inf, b]
            return self.base_dist.log_prob(value) - torch.log(self._cdf_at_high_right())
        if self.lower != -torch.inf and self.upper == torch.inf:
            # left bound -- [a, inf)
            sign = torch.where(self.base_dist.loc >= self.lower, 1.0, -1.0)
            return self.base_dist.log_prob(value) - torch.log(
                sign * (self._tail_prob_at_high_left() - self._tail_prob_at_low_left())
                )
        else:
            # two-sided bounds -- [a, b]
            return self.base_dist.log_prob(value) - self._log_diff_tail_probs_both()

    def _tail_prob_at_low_left(self):
        # if low < loc, returns cdf(low); otherwise returns 1 - cdf(low)
        loc = self.base_dist.loc
        sign = torch.where(loc >= self.lower, 1.0, -1.0)
        return self.base_dist.cdf(loc - sign * (loc - self.lower))

    def _tail_prob_at_high_left(self):
        # if low < loc, returns cdf(high) = 1; otherwise returns 1 - cdf(high) = 0
        return torch.where(self.lower <= self.base_dist.loc, 1.0, 0.0)

    def _cdf_at_high_right(self):
        return self.base_dist.cdf(torch.tensor(self.upper))

    def _tail_prob_at_low_both(self):
        # if low < loc, returns cdf(low); otherwise returns 1 - cdf(low)
        loc = self.base_dist.loc
        sign = torch.where(loc >= self.lower, 1.0, -1.0)
        return self.base_dist.cdf(loc - sign * (loc - self.lower))

    def _tail_prob_at_high_both(self):
        # if low < loc, returns cdf(high); otherwise returns 1 - cdf(high)
        loc = self.base_dist.loc
        sign = torch.where(loc >= self.lower, 1.0, -1.0)
        return self.base_dist.cdf(loc - sign * (loc - self.upper))

    def _log_diff_tail_probs_both(self):
        # use log_cdf method, if available, to avoid inf's in log_prob
        # fall back to cdf, if log_cdf not available
        log_cdf = getattr(self.base_dist, "log_cdf", None)
        if callable(log_cdf):
            return torch.logsumexp(
                a=torch.stack([log_cdf(self.upper), log_cdf(self.lower)], axis=-1),
                axis=-1,
                b=torch.Tensor([1, -1]),  # subtract low from high
            )

        else:
            loc = self.base_dist.loc
            sign = torch.where(loc >= self.lower, 1.0, -1.0)
            return torch.log(sign * (self._tail_prob_at_high_both() - self._tail_prob_at_low_both()))

    def _clamp_probs(self, probs):
        finfo = torch.finfo(torch.result_type(probs, 1.))
        return torch.clip(probs, min=finfo.tiny, max=1.0 - finfo.eps)

    def sample(self, sample_shape=()):
        shape = sample_shape + self.batch_shape + self.event_shape
        with torch.no_grad():
            u = torch.rand(shape)
            loc = self.base_dist.loc
            sign = torch.where(loc >= self.lower, 1.0, -1.0)

            if self.lower == -torch.inf and self.upper != torch.inf:
                return self.base_dist.icdf(u * self._cdf_at_high_right())
            if self.lower != -torch.inf and self.upper == torch.inf:
                return (1 - sign) * loc + sign * self.base_dist.icdf(
                    (1 - u) * self._tail_prob_at_low_left() + u * self._tail_prob_at_high_left()
                )
            else:
                return (1 - sign) * loc + sign * self.base_dist.icdf(
                    self._clamp_probs((1 - u) * self._tail_prob_at_low_both() + u * self._tail_prob_at_high_both())
                )


class TruncatedGamma(dist.TorchDistribution):
    arg_constraints = {'concentration': constraints.positive, 'rate': constraints.positive}
    support = constraints.positive

    def __init__(self, concentration, rate, lower_bound, upper_bound, validate_args=None):
        self.concentration = concentration
        self.rate = rate
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.base_dist = dist.Gamma(concentration, rate)
        super().__init__(self.base_dist.batch_shape, self.base_dist.event_shape, validate_args=validate_args)

    def sample(self, sample_shape=torch.Size()):
        samples = self.base_dist.sample(sample_shape)
        while (samples < self.lower_bound).any() or (samples > self.upper_bound).any():
            samples = torch.where(
                (samples < self.lower_bound) | (samples > self.upper_bound),
                self.base_dist.sample(sample_shape),
                samples
            )
        return samples

    def log_prob(self, value):
        log_prob = self.base_dist.log_prob(value)
        log_prob = torch.where(
            (value < self.lower_bound) | (value > self.upper_bound),
            torch.tensor(float('-inf')),
            log_prob
        )
        normalization_constant = (self.base_dist.cdf(self.upper_bound) - self.base_dist.cdf(self.lower_bound)).log()
        return log_prob - normalization_constant


# echidna.tools.model.py

from typing import Union

import pyro
import pyro.poutine as poutine
from pyro import distributions as dist
import torch.nn.functional as F

import torch

from echidna.tools.custom_dist import TruncatedNormal
from echidna.tools.utils import EchidnaConfig

class Echidna:
    def __init__(self, config: Union[EchidnaConfig, dict] = EchidnaConfig()):
        
        if isinstance(config, dict):
            config = EchidnaConfig(**config)
        self.config = config
        
        self.log_prob_scaler = 1.0 / (self.config.num_cells * self.config.num_genes)
        
        # set Echidna mode
        if self.config._is_multi:
            self.model = poutine.scale(self.model_mt, scale=self.log_prob_scaler)
            self.guide = poutine.scale(self.guide_mt, scale=self.log_prob_scaler)
        else:
            self.model = poutine.scale(self.model_st, scale=self.log_prob_scaler)
            self.guide = poutine.scale(self.guide_st, scale=self.log_prob_scaler)

        self.eta_posterior = None
        self.c_posterior = None
        self.cov_posterior = None
            
    def model_st(self, X, W, pi, z):
        library_size = X.sum(-1, keepdim=True) * 1e-5

        gene_plate = pyro.plate('G:genes', self.config.num_genes, dim=-1, device=self.config.device)
        cluster_plate = pyro.plate('K:clusters', self.config.num_clusters, dim=-2, device=self.config.device)

        clone_var_dist = dist.Gamma(1, 1).expand([self.config.num_clusters]).independent(1)
        scale = pyro.sample('scale', clone_var_dist)
        cov_dist = dist.LKJCholesky(self.config.num_clusters, self.config.lkj_concentration)
        cholesky_corr = pyro.sample('cholesky_corr', cov_dist)
        scale = scale[:, None] if self.config.inverse_gamma is False else 1/scale[:, None] # CHECK WITH MING
        cholesky_cov = cholesky_corr * torch.sqrt(scale) # INVERSE GAMMA OPTION

        # Sample eta
        with gene_plate:
            eta = pyro.sample(
                "eta",
                dist.MultivariateNormal(
                    torch.ones(self.config.num_clusters) * self.config.eta_mean_init
                    , scale_tril=cholesky_cov
                )
            )
            eta = F.softplus(eta).T

        # Sample W
        with gene_plate:
            pi = pyro.deterministic("pi", pi)
            W = pyro.sample('W', TruncatedNormal(pi @ eta, 0.05, lower=0.), obs=W)

        # Sample C
        with gene_plate:
            with cluster_plate:
                c = pyro.sample('c', dist.Gamma(1, 1/eta))

        # Sample X
        c_scale = c * torch.mean(eta,axis=1).repeat(self.config.num_genes,1).T
        z_tmp = pyro.deterministic("z", z.to(torch.int64))
        rate = c_scale[z_tmp] * library_size
        X = pyro.sample('X', dist.Poisson(rate).to_event(), obs=X)
        return X, W 

    def guide_st(self, X, W, pi, z):
        gene_plate = pyro.plate('G:genes', self.config.num_genes, dim=-1, device=self.config.device)
        cluster_plate = pyro.plate('K:clusters', self.config.num_clusters, dim=-2, device=self.config.device)

        q_eta_mean = pyro.param(
            'eta_mean', lambda:dist.MultivariateNormal(
                torch.ones(self.config.num_clusters) * self.config.eta_mean_init
                , torch.eye(self.config.num_clusters)
            ).sample([self.config.num_genes])
        )

        q_c_shape = pyro.param('c_shape', torch.ones(1, self.config.num_genes), constraint=dist.constraints.positive)

        shape = pyro.param('scale_shape', self.config.q_shape_rate_scaler * torch.ones(self.config.num_clusters),
                           constraint=dist.constraints.positive)
        rate = pyro.param('scale_rate', self.config.q_shape_rate_scaler * torch.ones(self.config.num_clusters),
                          constraint=dist.constraints.positive)
        q_clone_var = dist.Gamma(shape, rate).independent(1)
        q_scale = pyro.sample('scale', q_clone_var)

        corr_dim = self.config.num_clusters * (self.config.num_clusters - 1) // 2
        corr_loc = pyro.param("corr_loc", torch.zeros(corr_dim))
        corr_scale = pyro.param("corr_scale", torch.ones(corr_dim) * self.config.q_corr_init,
                                constraint=dist.constraints.positive)
        corr_cov = torch.diag(corr_scale)
        corr_dist = dist.MultivariateNormal(corr_loc, corr_cov)
        transformed_dist = dist.TransformedDistribution(corr_dist, dist.transforms.CorrCholeskyTransform())
        q_cholesky_corr = pyro.sample("cholesky_corr", transformed_dist)
        q_scale = q_scale[:, None] if self.config.inverse_gamma is False else 1/q_scale[:, None] # CHECK WITH MING
        q_cholesky_cov = q_cholesky_corr * torch.sqrt(q_scale) # INVERSE GAMMA OPTION

        with gene_plate:
            q_eta = pyro.sample('eta', dist.MultivariateNormal(q_eta_mean, scale_tril=q_cholesky_cov))
            q_eta = F.softplus(q_eta).T

        with gene_plate:
            with cluster_plate:
                pyro.sample("c", dist.Gamma(q_c_shape, 1/q_eta))

    def model_mt(self, X, W, pi, z):
        library_size = X.sum(-1, keepdim=True) * 1e-5
        num_timepoints = self.config.num_timepoints
        num_genes = self.config.num_genes
        num_clusters = self.config.num_clusters

        gene_plate = pyro.plate('G:genes', num_genes, dim=-1, device=self.config.device)
        cluster_plate = pyro.plate('K:clusters', num_clusters, dim=-2, device=self.config.device)

        # Eta covariance
        clone_var_dist = dist.Gamma(1, 1).expand([num_clusters]).independent(1)
        scale = pyro.sample('scale', clone_var_dist)
        cov_dist = dist.LKJCholesky(num_clusters, self.config.lkj_concentration)
        cholesky_corr = pyro.sample('cholesky_corr', cov_dist)
        scale = scale[:, None] if self.config.inverse_gamma is False else 1/scale[:, None] # CHECK WITH MING
        cholesky_cov = cholesky_corr * torch.sqrt(scale) # INVERSE GAMMA
        assert cholesky_cov.shape == (num_clusters, num_clusters) 
        # Sample eta
        with gene_plate:
            eta = pyro.sample('eta', dist.MultivariateNormal(self.config.eta_mean_init * torch.ones(num_clusters), scale_tril=cholesky_cov))
            eta = F.softplus(eta).T

        # Sample W per time point
        with gene_plate:
            with pyro.plate("timepoints_w", num_timepoints):
                pi = pyro.deterministic("pi", pi)
                mu_w = pi @ eta
                W = pyro.sample(f"W", TruncatedNormal(mu_w, 0.05, lower=0.), obs=W)

        # Sample c
        with gene_plate:
            with cluster_plate:
                with pyro.plate("timepoints_c", num_timepoints):
                    c = pyro.sample(f"c", dist.Gamma(1, 1/eta))

        for t in range(num_timepoints):
            c_scale = c[t, :, :] * torch.mean(eta,axis=-1).repeat(num_genes,1).T
            z_tmp = pyro.deterministic(f"z_{t}", z[t].to(torch.int64))
            rate = c_scale[z_tmp] * library_size[t]
            pyro.sample(f"X_{t}", dist.Poisson(rate).to_event(), obs=X[t])

    def guide_mt(self, X, W, pi, z):
        num_timepoints = self.config.num_timepoints
        num_genes = self.config.num_genes
        num_clusters = self.config.num_clusters

        gene_plate = pyro.plate('G:genes', num_genes, dim=-1, device=self.config.device)
        cluster_plate = pyro.plate('K:clusters', num_clusters, dim=-2, device=self.config.device)

        q_eta_mean = pyro.param('eta_mean',
                          lambda:dist.MultivariateNormal(torch.ones(num_clusters) * self.config.eta_mean_init,
                                                         torch.eye(num_clusters)).sample([num_genes]))
        q_c_shape = pyro.param('c_shape', torch.ones(num_timepoints, 1, num_genes), constraint=dist.constraints.positive)

        shape = pyro.param('scale_shape', self.config.q_shape_rate_scaler * torch.ones(num_clusters), constraint=dist.constraints.positive)
        rate = pyro.param('scale_rate', self.config.q_shape_rate_scaler * torch.ones(num_clusters), constraint=dist.constraints.positive)
        q_clone_var = dist.Gamma(shape, rate).independent(1)
        q_scale = pyro.sample('scale', q_clone_var)

        corr_dim = num_clusters * (num_clusters - 1) // 2
        corr_loc = pyro.param("corr_loc", torch.zeros(corr_dim))
        corr_scale = pyro.param("corr_scale", torch.ones(corr_dim) * self.config.q_corr_init,
                          constraint=dist.constraints.positive)
        corr_cov = torch.diag(corr_scale)
        corr_dist = dist.MultivariateNormal(corr_loc, corr_cov)
        transformed_dist = dist.TransformedDistribution(corr_dist, dist.transforms.CorrCholeskyTransform())
        q_cholesky_corr = pyro.sample("cholesky_corr", transformed_dist)
        q_scale = q_scale[:, None] if self.config.inverse_gamma is False else 1/q_scale[:, None] # CHECK WITH MING  
        q_cholesky_cov = q_cholesky_corr * torch.sqrt(q_scale) # INVERSE GAMMA OPTION 

        with gene_plate:
            q_eta = pyro.sample('eta', dist.MultivariateNormal(q_eta_mean, scale_tril=q_cholesky_cov))
            q_eta = F.softplus(q_eta).T

        with gene_plate:
            with cluster_plate:
                with pyro.plate("timepoints_c", num_timepoints):
                    pyro.sample("c", dist.Gamma(q_c_shape, 1/q_eta))



# echidna.tools.utils.py

import logging, os, shutil
import dataclasses
from dataclasses import dataclass, field

import pandas as pd
import scanpy as sc

from torch.cuda import is_available

from echidna.utils import (
    get_logger,
    ECHIDNA_GLOBALS,
    create_echidna_uns_key,
)

logger = get_logger(__name__)

@dataclass(unsafe_hash=True)
class EchidnaConfig:
    ## DATA PARAMETERS
    num_genes: int = None
    num_cells: int = None
    num_timepoints: int = None
    num_clusters: int = None
    
    timepoint_label: str="timepoint"
    counts_layer: str="counts"
    _is_multi: bool=None

    ## TRAINING PARAMETERS
    seed: int=42
    # max steps of SVI
    n_steps: int=10000
    # learning rate for Adam optimizer
    learning_rate: float=.1
    # % of training set to use for validation
    val_split: float=.1
    # cluster label to use in adata.obs
    clusters: str="pheno_louvain"
    # early stopping if patience > 0
    patience: int=None
    # gpus if available
    device: str="cuda" if is_available() else "cpu"
    # logging
    verbose: bool=True
    
    ## MODEL HYPERPARAMETERS
    # Use inverse gamma for noiser data
    inverse_gamma: bool=False
    # concentration parameter of LKJ. <1.0 more diag
    lkj_concentration: float = 1.0
    # scaler for the shape and rate parameters of covariance diag for variational inference
    q_shape_rate_scaler: float = 10.0
    # initialize the scale of variational correlation
    q_corr_init: float = 0.01
    # scaler for the covariance of variational correlation
    q_cov_scaler: float = 0.01
    # initial mean of eta
    eta_mean_init: float = 2.0
    # constant add to diag to ensure PD
    eps: float = 5e-3
    
    def to_dict(self):
        res = dataclasses.asdict(self)
        return res
    
    @classmethod
    def from_dict(cls, config_dict):
        config = cls()
        for key, value in config_dict.items():
            if hasattr(config, key):
                setattr(config, key, value)
        return config
    
    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

class EarlyStopping:
    """
    Borrowed from Decipher with author permission:
    Achille Nazaret, https://github.com/azizilab/decipher/blob/main/decipher/tools/utils.py
    
    Keeps track of when the loss does not improve after a given patience.
    Useful to stop training when the validation loss does not improve anymore.

    Parameters
    ----------
    patience : int
        How long to wait after the last validation loss improvement.

    Examples
    --------
    >>> n_epochs = 100
    >>> early_stopping = EarlyStopping(patience=5)
    >>> for epoch in range(n_epochs):
    >>>     # train
    >>>     validation_loss = ...
    >>>     if early_stopping(validation_loss):
    >>>         break
    """

    def __init__(self, patience=5):
        self.patience = patience
        self.counter = 0
        self.early_stop = False
        self.validation_loss_min = float("inf")

    def __call__(self, validation_loss):
        """Returns True if the training should stop."""
        if validation_loss < self.validation_loss_min:
            self.validation_loss_min = validation_loss
            self.counter = 0
        else:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True

        return self.early_stop

    def has_stopped(self):
        """Returns True if the stopping condition has been met."""
        return self.early_stop

def set_sort_order(adata, order):
    create_echidna_uns_key(adata)
    adata.uns["echidna"]["timepoint_order"] = {
        item: index for index, item in enumerate(order)
    }
    
def _custom_sort(items, order_dict):
    order_dict = {item: int(index) for item, index in order_dict.items()}
    default_order = len(order_dict)

    items_list = items.columns if isinstance(items, pd.DataFrame) else items
    sorted_list = sorted(items_list, key=lambda x: next((order_dict[sub] for sub in order_dict.keys() if sub in x), default_order))
    
    if isinstance(items, pd.DataFrame):
        return items.reindex(columns=sorted_list, fill_value=None)
    
    return sorted_list

def reset_echidna_memory():
    save_folder = ECHIDNA_GLOBALS["save_folder"]
    
    if os.path.exists(save_folder) and os.path.isdir(save_folder):
        shutil.rmtree(save_folder)
    logger.info("Cleared Echidna model saves.")



# echidna.tools.eval.py

import warnings
from anndata._core.views import ImplicitModificationWarning

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.cluster.hierarchy import dendrogram, linkage, cophenet
from scipy.spatial.distance import squareform
from scipy.spatial.distance import pdist
from scipy.ndimage import gaussian_filter1d

import torch
import pyro
import pyro.distributions as dist

from echidna.tools.custom_dist import TruncatedNormal
from echidna.tools.housekeeping import load_model
from echidna.tools.utils import _custom_sort
from echidna.tools.data import create_z_pi, filter_low_var_genes
from echidna.utils import get_logger

logger = get_logger(__name__)

def sample(adata, variable, **kwargs):
    if isinstance(variable, str):
        variable = [variable]
    elif not isinstance(variable, list):
        raise ValueError("`variable` must be either a string or a list.")
    
    allowed_vars = {"X", "W", "c", "cov", "eta"}
    
    invalid_vars = set(variable) - allowed_vars
    if invalid_vars:
        raise ValueError(
            f"The following `variable` values are not allowed: {invalid_vars}. "
            "Allowed values are (\"X\", \"W\", \"c\", \"eta\", \"cov\")."
        )
    
    sample_funcs = {
        "X": sample_X,
        "W": sample_W,
        "c": sample_c,
        "eta": sample_eta,
        "cov": sample_cov,
    }
    
    samples = [sample_funcs[v](adata, **kwargs) for v in variable]
    if len(samples) == 1: return samples[0]
    return samples

def sample_X(adata, num_cells=None, return_z=False):
    """Sample X given posterior estimates of c and eta
    """
    config = adata.uns["echidna"]["config"]

    echidna = load_model(adata)
    
    # num_cells cleaning and checks
    num_cells = int(config["num_cells"]) if num_cells is None else num_cells
    num_cells = num_cells[0] if isinstance(num_cells, (tuple, list, torch.Tensor, np.ndarray)) else num_cells
    if not isinstance(num_cells, int): raise ValueError("`num_cells` must be of type int.")
    
    # Check num_cells is within bounds
    ## Get the number of cells in each timepoint, which should
    ## be even across TP via the discard filter
    discard_filter = adata.obs["echidna_split"] != "discard"
    timepoint_label = config["timepoint_label"]
    arbitrary_tp_filter = adata.obs.loc[:, timepoint_label] == adata.obs.loc[:, timepoint_label].iloc[0]
    max_cells = adata.obs[discard_filter & arbitrary_tp_filter].shape[0]
    if num_cells > max_cells:
        raise ValueError(f"`num_cells` must be less than {max_cells}.")
        
    rng = np.random.default_rng(int(config["seed"]))
    
    z = []
    library_size = []
    timepoints = np.unique(adata.obs[config["timepoint_label"]])
    if "timepoint_order" in adata.uns["echidna"]: 
        timepoints = _custom_sort(timepoints, adata.uns["echidna"]["timepoint_order"])
    if "total_counts" not in adata.obs:
        adata.obs["total_counts"] = adata.layers[config["counts_layer"]].sum(-1)
        logger.info(
            "Added `.obs['total_counts']` : Library size, sum of counts"
            " across genes for each cell."
        )
    for i, t in enumerate(timepoints):
        timepoint_filter = adata.obs[config["timepoint_label"]]==t
        adata_tmp = adata[timepoint_filter, adata.var["echidna_matched_genes"]]
        cell_index = range(num_cells)
        z.append(
            torch.tensor(adata_tmp.obs.loc[
                adata_tmp.obs.index[cell_index], config["clusters"]
            ].values, dtype=torch.int64)
        )
        library_size.append(
            torch.tensor(
                adata_tmp.obs.loc[:, "total_counts"].iloc[cell_index].values,
                dtype=torch.float32,
            )
        )
    
    library_size = torch.stack(library_size) * 1e-5
    z = torch.stack(z)
    c = echidna.c_posterior
    if bool(config["_is_multi"]) is False: c = c.unsqueeze(0)
    eta = echidna.eta_posterior

    mean_scale = torch.mean(eta, axis=-1).repeat(int(config["num_genes"]),1).T
    c_scaled = c * mean_scale.unsqueeze(0)

    X_sample = []
    for t in range(int(config["num_timepoints"])):
        rate = c_scaled[t, z[t], :] * library_size[t,:].view(-1,1)
        X = dist.Poisson(rate).sample()
        X_sample.append(X)
    X_sample = torch.stack(X_sample, dim=0).squeeze()

    del echidna
    if return_z:
        return X_sample, z
    return X_sample

def sample_W(adata, num_samples=(1,)):
    """Sample W given posterior estimates of eta
    """
    config = adata.uns["echidna"]["config"]
    echidna = load_model(adata)
    
    pi, _ = create_z_pi(adata, config)
    
    W = TruncatedNormal(pi @ echidna.eta_posterior, 0.05, lower=0).sample(num_samples)
    W = W.squeeze()
    del echidna
    return W

def sample_c(adata, num_samples=(1,)) -> torch.Tensor:
    """Sample C from posterior and selec a target cluster for
    a given time point.
    """
    config = adata.uns["echidna"]["config"]
    echidna = load_model(adata)
    
    eta = dist.MultivariateNormal(
        echidna.eta_posterior.expand(config["num_clusters"], -1).T,
        covariance_matrix=echidna.cov_posterior
    ).sample()
    c_sample = dist.Gamma(pyro.param("c_shape"), 1/eta)
    c_sample = c_sample.sample(num_samples).squeeze()
    del echidna
    return c_sample

def sample_eta(adata, num_samples=1000):
    """Sample eta from posterior
    """
    config = adata.uns["echidna"]["config"]
    echidna = load_model(adata)
    
    eta_posterior = dist.MultivariateNormal(
        echidna.eta_posterior.T,
        covariance_matrix=echidna.cov_posterior
    )
    eta_samples = eta_posterior.sample([num_samples,]).squeeze()
    del echidna
    return torch.nn.functional.softplus(eta_samples)

def sample_cov(adata, num_samples=(1,)):
    """Sample a covariance matrix from the posterior distribution.

    Parameters
    ----------
    adata : AnnData
        The annotated data matrix.
    num_samples : int, optional
        The number of samples to draw from the posterior distribution.
        By default, draw a single sample.

    Returns
    -------
    covariance : torch.Tensor
        A covariance matrix sampled from the posterior distribution.
        The covariance matrix is of shape `(num_genes, num_genes)`.
    """
    model = load_model(adata)
    config = model.config
    
    # Sample a Cholesky decomposition of the covariance matrix
    correlation_loc = pyro.param("corr_loc")
    correlation_scale = pyro.param("corr_scale")
    correlation_dist = dist.MultivariateNormal(correlation_loc, torch.diag(correlation_scale))
    transformed_dist = dist.TransformedDistribution(correlation_dist, dist.transforms.CorrCholeskyTransform())
    chol_samples = transformed_dist.sample(num_samples).squeeze()
    
    # Sample the scale for the covariance matrix
    scale_shape = pyro.param("scale_shape")
    scale_rate = pyro.param("scale_rate")
    scale = scale_shape / scale_rate
    
    # Compute the covariance matrix
    if not config.inverse_gamma:
        scale = scale[:, None]
    else:
        scale = 1 / scale[:, None]
    covariance = chol_samples * torch.sqrt(scale)
    covariance = covariance @ covariance.T
    
    return covariance

def normalize_cov(cov_matrix):
    """Normalize a covariance matrix to a correlation matrix.

    Parameters
    ----------
    cov_matrix : torch.Tensor, shape (n_features, n_features)
        Covariance matrix.

    Returns
    -------
    corr_matrix : torch.Tensor, shape (n_features, n_features)
        Correlation matrix.
    """
    std_dev = torch.sqrt(torch.diag(cov_matrix))
    outer_std_dev = std_dev[:, None] * std_dev[None, :]
    corr_matrix = cov_matrix / outer_std_dev
    corr_matrix.diagonal(0).fill_(1)
    return corr_matrix

def mahalanobis_distance_matrix(eta):
    cov_matrix = torch.cov(eta)
    cov_inv = torch.linalg.inv(cov_matrix)
    
    num_vectors = eta.size(1)
    distance_matrix = torch.zeros((num_vectors, num_vectors), device=eta.device)
    
    for i in range(num_vectors):
        diff_i = eta.T - eta.T[i]  # Broadcasting difference for row i
        distance_matrix[i] = torch.sqrt((diff_i @ cov_inv @ diff_i.T).diag())
    
    return distance_matrix

def distance_matrix_helper(eta, similarity_metric):
    if similarity_metric == "smoothed_corr":
        eta = eta.cpu().detach().numpy()
        eta = gaussian_filter1d(eta, sigma=6, axis=1, radius=8)
        return pdist(eta, metric="correlation")
    elif similarity_metric == "cov":
        return torch.cov(eta).cpu().detach().numpy()
    elif similarity_metric == "corr":
        eta_corr = torch.corrcoef(eta).cpu().detach().numpy()
        return 1 - eta_corr
    elif similarity_metric == "mahalanobis":
        return mahalanobis_distance_matrix(eta).cpu().detach().numpy()
    else:
        raise ValueError(
            "Invalid similarity_metric. Use `smoothed_corr`, `cov`, `corr`."
        )

def eta_tree_elbow_thresholding(
    eta,
    similarity_metric,
    plot_dendrogram: bool=False,
    plot_elbow: bool=False
):
    dist_mat = distance_matrix_helper(eta, similarity_metric)
    link_metric = "ward" if similarity_metric == "smoothed_corr" else "average"
    Z = linkage(dist_mat, link_metric)
    distance = Z[:, 2]
    differences = np.diff(distance)
    knee_point = np.argmax(differences)
    threshold = distance[knee_point]

    if plot_elbow:
        fig, ax = plt.subplots()
        ax.plot(range(1, len(differences) + 1), differences, marker="o")
        ax.axvline(x=knee_point + 1, linestyle="--", label="ELBOW threshold", color="red")
        ax.legend()
        ax.set_xlabel("Merge Step")
        ax.set_ylabel("Difference in Distance")
        ax.set_title("Elbow Method")
        return fig
    elif plot_dendrogram:
        fig, ax = plt.subplots()
        dendrogram(Z, color_threshold=threshold, no_plot=False, ax=ax)
        ax.set_title("Echidna Clusters Hierarchy")
        return fig
    else: 
        logger.info(f"Dendrogram knee point: {knee_point + 1}")
        logger.info(f"Dendrogram threshold: {threshold:.4f}")
        return dendrogram(Z, color_threshold=threshold, no_plot=True)

def eta_tree_cophenetic_thresholding(
    eta, 
    similarity_metric,
    frac=0.7, 
    dist_matrix=False,
    plot_dendrogram: bool=False,
):
    dist_mat = distance_matrix_helper(eta, similarity_metric)
    link_metric = "ward" if similarity_metric == "smoothed_corr" else "average"
    if dist_matrix:
        Z = linkage(squareform(dist_mat), link_metric)
    else:
        Z = linkage(dist_mat, link_metric)
    coph_distances = cophenet(Z)
    max_coph_distance = np.max(coph_distances)
    threshold = frac * max_coph_distance
    
    if not plot_dendrogram:
        logger.info(f"Dendrogram threshold: {threshold:.4f}")
        return dendrogram(Z, color_threshold=threshold, no_plot=True)
    else:
        dendrogram(Z, color_threshold=threshold, no_plot=False)

def eta_tree(
    eta, 
    similarity_metric,
    thres: float, 
    plot_dendrogram: bool=False
):
    dist_mat = distance_matrix_helper(eta, similarity_metric)
    link_metric = "ward" if similarity_metric == "smoothed_corr" else "average"
    Z = linkage(dist_mat, link_metric)
    if not plot_dendrogram:
        return dendrogram(Z, color_threshold=thres, no_plot=True)
    else:
        dendrogram(Z, color_threshold=thres, no_plot=False)

def assign_clones(dn, adata):
    """Assign clones based on the covariance tree for each cell.

    Parameters
    ----------
    dn : dict
        A dictionary containing the leaves color list and leaves information.
    adata : sc.AnnData
        Annotated data matrix.
    """

    if "echidna" not in adata.uns:
        raise ValueError(
            "No echidna model has been saved for this AnnData object."
        )
        
    clst = dn.get('leaves_color_list')
    cluster_count = np.unique(clst).shape[0]
    for i in range(len(clst)):
        if clst[i] == 'C0':
            clst[i] = f'C{cluster_count}'
            cluster_count += 1
    keys = dn.get('leaves')
    color_dict = pd.DataFrame(clst)
    color_dict.columns = ['color']
    color_dict.index = keys
    cluster_label = adata.uns["echidna"]["config"]["clusters"]
    hier_colors = [color_dict.loc[int(i)][0] for i in adata.obs[cluster_label]]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ImplicitModificationWarning)
        adata.obs["echidna_clones"] = pd.Series(
            hier_colors,
            index=adata.obs.index,
            dtype="category",
        )
        
def echidna_clones(
    adata,
    method: str="elbow",
    metric: str=None,
    threshold: float=0.,
):
    echidna = load_model(adata)
    metric = "smoothed_corr" if metric is None else metric
    adata.uns["echidna"]["save_data"]["dendrogram_metric"] = metric
    
    # If a threshold is set, use that threshold
    if threshold > 0.:
        method = "manual"
    
    adata.uns["echidna"]["save_data"]["dendrogram_method"] = method
    
    if method == "elbow":
        dn = eta_tree_elbow_thresholding(
            echidna.eta_posterior, similarity_metric=metric
        )
    elif method == "cophenetic":
        dn = eta_tree_cophenetic_thresholding(
            echidna.eta_posterior, similarity_metric=metric
        )
    else:
        adata.uns["echidna"]["save_data"]["threshold"] = threshold
        if threshold == 0.:
            logger.warning(
                "If not using `elbow` or `cophenetic` method, "
                "you must set `threshold` manually. Default `threshold=0`."
            )
        dn = eta_tree(
            echidna.eta_posterior, similarity_metric=metric, thres=threshold
        )
    
    assign_clones(dn, adata)
    logger.info(
        "Added `.obs['echidna_clones']`: the learned clones from eta."
    )
    del echidna

def echidna_status(adata, threshold: float = 0.):
    
    if "echidna_clones" not in adata.obs.columns:
        raise ValueError("Must run `ec.tl.echidna_clones` first.")
    if "echidna_status" in adata.obs.columns:
        adata.obs.drop(columns="echidna_status", inplace=True)

    timepoint_label = adata.uns["echidna"]["config"]["timepoint_label"]

    # Group by echidna_clones and timepoint_label to get
    # num cells in each clone/timepoint
    clones_gb = adata.obs.groupby(
        by=["echidna_clones", timepoint_label]
    ).size().reset_index(name="clone_tp_counts")

    # Group by timepoint_label to get total counts at each timepoint
    tp_gb = clones_gb.groupby(
        timepoint_label
    )["clone_tp_counts"].sum().reset_index(name="tp_counts")

    # Merge the data to get timepoint counts for each row in clone/timepoint count
    clones_gb = clones_gb.merge(
        tp_gb,
        left_on=timepoint_label,
        right_on=timepoint_label,
    )

    # Calculate the ratio of clone/timepoint counts to timepoint total counts (eg "pre / pre_total")
    clones_gb["tp_ratio"] = clones_gb["clone_tp_counts"] / clones_gb["tp_counts"]
    
    # Map integer order to timepoints for sorting
    clones_gb["timepoint_order"] = clones_gb[timepoint_label].map(
        adata.uns["echidna"]["timepoint_order"]
    )

    # Sort by echidna_clones and timepoint_order, ascending=True since we get t+1/t
    clones_gb = clones_gb.sort_values(by=["echidna_clones", "timepoint_order"])

    # Define a function to calculate the status_score and status for each clone
    def calculate_status(group):
        # sort again to be sure the shift is done correctly
        group = group.sort_values(by="timepoint_order")
        # Generalized version of "ons / pres". Shift lets us calculate t+1/t
        group["status_score"] = np.log2(group["tp_ratio"] / group["tp_ratio"].shift(-1)).ffill()
        group["echidna_status"] = np.where(
            group["status_score"] > threshold, "growing", "stable"
        )
        group["echidna_status"] = np.where(
            group["status_score"] < -threshold, "shrinking", group["echidna_status"]
        )
        group["echidna_status"] = pd.Categorical(group["echidna_status"])
        return group

    # Apply the function to each group of echidna_clones
    result = clones_gb.groupby("echidna_clones").apply(calculate_status).reset_index(drop=True)

    # Merge the status back into adata.obs
    adata.obs = adata.obs.merge(
        result.loc[:, ["echidna_clones", timepoint_label, "echidna_status"]],
        on=["echidna_clones", timepoint_label],
        how="left"
    ).set_index(adata.obs.index)
    
    adata.obs["echidna_status"] = adata.obs["echidna_status"].astype("category")

    logger.info(
        "Added `echidna_status` to `.obs` : "
        "Labeling Echidna clones as growing, shrinking, or stable."
    )


# echidna.plot.__init__.py

from .ppc import plate_model, ppc
from .post import (
    dendrogram,
    echidna,
    plot_cnv,
    plot_eta,
    plot_gene_dosage
)
from .utils import activate_plot_settings

__all__ = [
    "plate_model",
    "ppc",
    "dendrogram",
    "echidna",
    "plot_cnv",
    "plot_eta",
    "plot_gene_dosage",
    "activate_plot_settings",
]


# echidna.tools.post.py

import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc
import numpy as np
import pandas as pd
import os
from typing import Union, List
import torch
from scipy.ndimage import gaussian_filter1d

from echidna.tools.eval import (
    eta_tree,
    eta_tree_elbow_thresholding,
    eta_tree_cophenetic_thresholding,
)
from echidna.tools.infer_gd import gene_dosage_effect, cnv_results
from echidna.tools.housekeeping import load_model
from echidna.tools.data import sort_chromosomes, filter_low_var_genes
from echidna.plot.utils import save_figure, activate_plot_settings
from echidna.utils import get_logger

logger = get_logger(__name__)

def plot_eta(adata, filename: str=None):
    cluster_label = adata.uns["echidna"]["config"]["clusters"]
    
    clust_order = []
    for i in np.unique(adata.obs["echidna_clones"]):
        clust_order.extend(
            adata.obs.loc[
                adata.obs["echidna_clones"]==i,
                cluster_label,
            ].unique()
        )
    eta = cnv_results(adata)
    num_clusters = adata.obs[cluster_label].nunique()
    eta_vals = eta[[f"echidna_clone_{x}" for x in range(num_clusters)]]
    
    eta_vals.columns = eta_vals.columns.str.extract(r'(\d+)$')[0].astype(int)
    eta_vals = eta_vals[clust_order]
    
    fig, ax = plt.subplots(figsize=(25, 5))
    smoothed_eta = gaussian_filter1d(eta_vals.T, sigma=6, axis=1, radius=8)
    sns.heatmap(pd.DataFrame(smoothed_eta, index=eta_vals.columns, columns=eta_vals.index), cmap='coolwarm', vmin=-2, vmax=2, ax=ax)
    chrom_counts = sort_chromosomes(
        eta["chrom"].value_counts()
    ).cumsum()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ticks = [(chrom_counts[i-1] + chrom_counts[i])/2 if i != 0 else chrom_counts[i]/2 for i in range(len(chrom_counts))]
    ax.set_xticks(ticks)
    ax.set_xticklabels(chrom_counts.index, rotation=30)
    
    for x in chrom_counts:
        ax.axvline(x=x, color="k", linestyle="--", linewidth=1.5)

def plot_cnv(adata, c: str=None, filename: str=None):
    c = "all" if c is None else c
    
    # Retrieve save data from tools functions
    if "echi_cnv" not in adata.uns["echidna"]["save_data"]:
        raise ValueError("Must run `ec.tl.echi_cnv` first.")
    file_save_path = adata.uns["echidna"]["save_data"]["echi_cnv"]
    neutral_save_path = adata.uns["echidna"]["save_data"]["gmm_neutrals"]
    if not os.path.exists(file_save_path):
        raise ValueError(
            "Saved results not found. Run `ec.tl.echi_cnv` first."
        )
    eta_genome_merge = pd.read_csv(file_save_path)
    neutral_states = pd.read_csv(
        neutral_save_path,
        index_col="eta_column_label",
    )
    cluster_label = adata.uns["echidna"]["config"]["clusters"]
    num_clusters = adata.obs[cluster_label].nunique()
    
    eta_genome_merge["chrom"] = eta_genome_merge["chrom"].str.extract(r"^(chr[0-9XY]+)")[0]
    
    cols = []
    for i in [f"echidna_clone_{x}" for x in range(num_clusters)]:
        cols.append(i)
        eta_genome_merge[i] -= neutral_states.loc[i, "neutral_value_mean"].item()
    
    chrom_counts = sort_chromosomes(
        eta_genome_merge["chrom"].value_counts()
    ).cumsum()
    
    clust_order = []
    for i in np.unique(adata.obs["echidna_clones"]):
        clust_order.extend(
            ["echidna_clone_" + str(x) for x in adata.obs.loc[
                adata.obs["echidna_clones"] == i,
                cluster_label,
            ].unique()]
        )
    
    activate_plot_settings()
    if c != "all":
        if f"echidna_clone_{c}" not in eta_genome_merge.columns:
            raise ValueError(f"Specified cluster `echidna_clone_{c}` not found.")
        vals = eta_genome_merge.loc[:, f"echidna_clone_{c}"]
        states = eta_genome_merge.loc[:, f"states_echidna_clone_{c}"]
        _plot_cnv_helper(
            vals,
            states,
            chrom_counts.values,
            chrom_counts.index,
            title=f"Echidna Clone {c} CNV",
            filename=filename,
        )  
    elif c == "all":
        fig, ax = plt.subplots(figsize=(20,5), nrows=1, ncols=1)
        
        eta_genome_merge = eta_genome_merge[clust_order]
        eta_genome_merge.columns = eta_genome_merge.columns.str.extract(r'(\d+)$')[0].astype(int)
        sns.heatmap(eta_genome_merge.T, cmap="bwr", ax=ax, vmin=-2, vmax=2)
        ax.set_xlabel("")
        ax.set_ylabel("")
        # Set the x-axis ticks and labels
        ticks = [(chrom_counts[i-1] + chrom_counts[i])/2 if i != 0 else chrom_counts[i]/2 for i in range(len(chrom_counts))]
        ax.set_xticks(ticks)
        ax.set_xticklabels(chrom_counts.index, rotation=30)

        # Draw vertical lines at each chromosome boundary
        for x in chrom_counts:
            ax.axvline(x=x, color="k", linestyle="--", linewidth=1.5)
         
        # fig, axes = plt.subplots(num_clusters, 1, figsize=(25, 7 * num_clusters))

        # for i in range(num_clusters):
        #     vals = band_means_states.loc[:, f"echidna_clone_{i}"]
        #     states = band_means_states.loc[:, f"states_echidna_clone_{i}"]
        #     _plot_cnv_helper(
        #         vals,
        #         states,
        #         chrom_counts.values,
        #         chrom_counts.index,
        #         ax=axes[i],
        #         title=f"Echidna Clone {i} CNV",
        #         filename=None,
        #     )
        
        if filename: fig.savefig(filename, format="png")

def _plot_cnv_helper(vals, states, chrom_coords, chroms, ax=None, title=None, filename=None):
    """Plot the CNV states along the genome.

    Parameters
    ----------
        vals : list/np.ndarray
            List of ordered copy number values (from bin_by_bands function)
        states : list/np.ndarray
            List of CN state calls from the HMM (get_states function)
        chrom_coords : list
            list of coordinates of the end of each chromosome (from bin_by_bands)
        chroms : list
            Ordered unique list of chromosome names
        title : str (optional)
            Title to label the plot
        filename : str (optional)
            Name of the file to save the plot
    """
    df = pd.DataFrame({
        "x": np.arange(len(vals)),
        "vals": vals,
        "states": states,
    })
    
    color_map = {"neut": "grey", "amp": "red", "del": "blue"}
    df["color"] = df["states"].map(color_map)
    
    if ax is None:
        fig, ax = plt.subplots(figsize=(25, 5))
    
    # for i, row in df.iterrows():
        # ax.axvline(x=row["x"], color=row["color"], linestyle="-", alpha=0.3, linewidth=1)
    
    sns.scatterplot(x="x", y="vals", hue="states", palette=color_map, data=df, legend=False, s=80, ax=ax)
    
    # Set the x-axis ticks and labels
    ticks = [(chrom_coords[i-1] + chrom_coords[i])/2 if i != 0 else chrom_coords[i]/2 for i in range(len(chrom_coords))]
    ax.set_xticks(ticks)
    ax.set_xticklabels(chroms, rotation=30)
    
    # Draw vertical lines at each chromosome boundary
    for x in chrom_coords:
        ax.axvline(x=x, color="k", linestyle="--", linewidth=1.5)
        
    ax.set_xlabel("Bands")
    ax.set_ylabel("CN")
    
    ax.grid(axis="x")
    
    if title:
        ax.set_title(title)
    if filename:
        ax.figure.savefig(filename, format="png")
        
def plot_gene_dosage(
    adata,
    clusters: Union[int, List[int]]=None,
    timepoints: Union[int, List[int]]=None,
    quantile: float=.8,
    var_threshold: float=None,
    filename: str=None,
) -> None:
    model = load_model(adata)
    
    if "gene_dosage" not in adata.uns["echidna"]["save_data"]:
        gene_dosage = gene_dosage_effect(adata)
    else:
        gene_dosage_cache = adata.uns["echidna"]["save_data"]["gene_dosage"]
        if not os.path.exists(gene_dosage_cache):
            raise ValueError(
                "Saved results not found. Run `ec.tl.gene_dosage_effect` first."
            )
        gene_dosage = torch.load(gene_dosage_cache).to(model.config.device)
    
    # Ensure clusters is a list of integers
    if clusters is not None:
        if isinstance(clusters, int):
            clusters = [clusters]
        elif not all(isinstance(cl, int) for cl in clusters):
            raise ValueError("Clusters must be an integer or a list of integers.")
        
        if not all(0 <= cl < gene_dosage.shape[-1] for cl in clusters):
            raise ValueError(
                f"Invalid cluster values: {clusters}. "
                f"Clusters must be integers in the range [0, {gene_dosage.shape[-1] - 1}]."
            )
    else:
        clusters = list(range(0, gene_dosage.shape[-1]))

    # Ensure timepoints is a list of integers
    if timepoints is not None:
        if isinstance(timepoints, int):
            timepoints = [timepoints]
        elif not all(isinstance(tp, int) for tp in timepoints):
            raise ValueError("Timepoints must be an integer or a list of integers.")
        
        if not all(0 <= tp < gene_dosage.shape[-2] for tp in timepoints):
            raise ValueError(
                f"Invalid timepoint values: {timepoints}. "
                f"Timepoints must be integers in the range [0, {gene_dosage.shape[-2] - 1}]."
            )
    else:
        timepoints = list(range(0, gene_dosage.shape[-2]))
    
    neutral_save_path = adata.uns["echidna"]["save_data"]["gmm_neutrals"]
    neutral_states = pd.read_csv(
        neutral_save_path,
        index_col="eta_column_label",
    )
    eta = model.eta_posterior
    echidna_matched_genes = adata[
        :, adata.var["echidna_matched_genes"]
    ].var.index

    eta = pd.DataFrame(
        model.eta_posterior.T.cpu().detach().numpy(),
        index=echidna_matched_genes,
        # columns=[f"echidna_clone_{i}" for i in range(num_clusters)]
    )
    
    filter_genes = filter_low_var_genes(
        adata[:, adata.var["echidna_matched_genes"]].copy(),
        quantile=quantile,
        var_threshold=var_threshold,
        indices=True,
    )

    filter_genes_tensor = torch.tensor(filter_genes, device=model.config.device)
    filter_genes_indices = torch.nonzero(filter_genes_tensor, as_tuple=False).squeeze()
    eta = eta[filter_genes]
    
    n_rows = len(clusters)
    n_cols = len(timepoints)
    
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(5 * n_cols, 5 * n_rows), squeeze=False)
    
    # Loop over each cluster and timepoint to create subplots
    for i, tp in enumerate(timepoints):
        gene_dosage_df = pd.DataFrame(
                gene_dosage[filter_genes_indices, tp, :].abs().cpu().detach().numpy(),
                index=eta.index,#adata[:, adata.var["echidna_matched_genes"]].var.index,
        )
        for j, cl in enumerate(clusters):

            ax = axes[j][i]
            
            sns.scatterplot(
                x=eta[cl] - neutral_states.loc[f"echidna_clone_{cl}", "neutral_value_mean"],
                y=gene_dosage_df[cl],
                ax=ax
            )
            ax.set_title(f"Cluster {cl}, Timepoint {tp}")
            ax.set_xlabel("Eta Shifted")
            ax.set_ylabel("Gene Dosage Effect")

    plt.tight_layout()
    if filename:
        plt.savefig(filename)
    plt.show()
    
def plot_loss(losses: list, label: str, log_loss: bool=False):
    activate_plot_settings()
    fig,ax = plt.subplots(1, 2, figsize=(12,4), sharey=False)
    if log_loss:
        sns.lineplot(np.log10(losses), ax=ax[0])
    else:
        sns.lineplot(losses, ax=ax[0])
    ax[0].set_title(f"{label} loss")
    ax[0].set_xlabel("steps")

    sns.lineplot(np.diff(losses), ax=ax[1])
    ax[1].set_title("step delta")
    ax[1].set_xlabel("steps")

    plt.show()

def dendrogram(adata, elbow: bool=False, filepath: str=None):
    activate_plot_settings()
    echidna = load_model(adata)
    try:
        method = adata.uns["echidna"]["save_data"]["dendrogram_method"]
        metric = adata.uns["echidna"]["save_data"]["dendrogram_metric"]
    except Exception as e:
        logger.error(f"Must run `ec.tl.echidna_clones` first. {e}")
        return

    if elbow and method != "elbow":
        logger.warning("`elbow=True` only applies to `method=\"elbow\"`.")

    if method == "elbow":
        fig = eta_tree_elbow_thresholding(
            echidna.eta_posterior,
            similarity_metric=metric,
            plot_dendrogram=not elbow,
            plot_elbow=elbow,
        )
    elif method == "cophenetic":
        fig = eta_tree_cophenetic_thresholding(
            echidna.eta_posterior,
            similarity_metric=metric,
            plot_dendrogram=True,
        )
    else:
        fig = eta_tree(
            echidna.eta_posterior,
            similarity_metric=metric,
            thres=adata.uns["echidna"]["save_data"]["threshold"],
            plot_dendrogram=True,
        )

    if filepath: save_figure(fig, filepath)
    del echidna

def echidna(
    adata,
    color=["echidna_clones"],
    basis="X_umap",
    filepath=None,
    return_fig=False,
    **kwargs,
):
    """
    Parameters
    ----------
    adata : sc.AnnData
        The annotated data matrix.
    basis : str, default "X_umap"
        The basis to use for the plot.
    **kwargs : dict, optional
        Additional arguments passed to `sc.pl.embedding`.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        The matplotlib figure.
    """
    activate_plot_settings()
    
    fig = sc.pl.embedding(
        adata,
        basis=basis,
        color=color,
        color_map="cool_r",
        frameon=True,
        show=True,
        sort_order=True,
        return_fig=True,
        **kwargs,
    )
    if filepath: save_figure(fig, filepath)
    if return_fig: return fig



# echidna.utils.py

ECHIDNA_GLOBALS = dict()
ECHIDNA_GLOBALS["save_folder"] = "./_echidna_models/"

def create_echidna_uns_key(adata):
    """
    Create the `echidna` uns key if it doesn't exist.

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    """
    if "echidna" not in adata.uns:
        adata.uns["echidna"] = dict()
    if "save_data" not in adata.uns["echidna"]:
        adata.uns["echidna"]["save_data"] = dict()
    if "timepoint_order" not in adata.uns["echidna"]:
        adata.uns["echidna"]["timepoint_order"] = dict()
        
def get_logger(name):
    import logging
    logger = logging.getLogger(name)
    logging.basicConfig(
        format="%(asctime)s | %(levelname)s : %(message)s",
        level=logging.INFO,
    )
    return logger



# echidna.plot.utils.py

import matplotlib as mpl
import matplotlib.pyplot as plt
import scanpy as sc
import seaborn as sns
import warnings

def save_figure(fig, filepath, dpi=300, bbox_inches='tight'):
    """
    Save a Matplotlib figure to a file.

    Parameters:
    fig : matplotlib.figure.Figure
        The figure object to be saved.
    filepath : str
        The file path where the figure will be saved.
    dpi : int, optional
        The resolution of the saved figure in dots per inch (default is 300).
    bbox_inches : str, optional
        Bounding box in inches: 'tight' to include all elements (default is 'tight').
    """
    fig.savefig(filepath, dpi=dpi, bbox_inches=bbox_inches)
    plt.close(fig)

def is_notebook():
    """
    Check if the script is running in a Jupyter notebook.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False
    except NameError:
        return False

def activate_plot_settings():
    """Activate journal quality settings for plotting.
    
    Modified from Decipher with author permission:
    Achille Nazaret, https://github.com/azizilab/decipher/blob/main/decipher/plot/utils.py

    It is recommended for high quality figures while keeping the file size small.
    It is recommended if the figures are to be edited in Adobe Illustrator.
    """
    if is_notebook():        
        sc.settings.set_figure_params(dpi_save=200, vector_friendly=True, fontsize=12)
        warnings.filterwarnings("ignore", category=UserWarning, message="No data for colormapping provided via 'c'")
        # Set Matplotlib font types for vector graphics
        mpl.rcParams["pdf.fonttype"] = 42
        mpl.rcParams["ps.fonttype"] = 42
        mpl.rcParams["figure.figsize"] = [6, 4]
        mpl.rcParams["savefig.dpi"] = 200
    else:
        mpl.rcParams["savefig.dpi"] = 400
        
    sns.set_theme(style="white", rc={"grid.color": ".6", "grid.linestyle": ":"})
    sns.set_context("paper", font_scale=1.)
    
    mpl.rcParams["axes.grid"] = True
    mpl.rcParams["figure.autolayout"] = False
    mpl.rcParams["legend.frameon"] = True



# echidna.tools.housekeeping.py

from datetime import datetime
import logging
import os, gc

import numpy as np

import torch
import torch.nn.functional as F

import pyro
import pyro.distributions as dist

from echidna.tools.utils import EchidnaConfig
from echidna.tools.model import Echidna
from echidna.utils import (
    ECHIDNA_GLOBALS,
    create_echidna_uns_key,
    get_logger,
)

logger = get_logger(__name__)

def set_posteriors(echidna, data):
    """
    Set model posteriors after training
    """
    echidna.eta_posterior = eta_posterior_estimates(echidna, data)
    echidna.c_posterior = c_posterior_estimates(eta=echidna.eta_posterior, mt=echidna.config._is_multi)
    echidna.cov_posterior = cov_posterior_estimate(inverse_gamma=echidna.config.inverse_gamma)
    echidna.corr_posterior = normalize_cov(echidna.cov_posterior)
    echidna.library_size = data[0].sum(-1, keepdim=True) * 1e-5
    return echidna

def get_learned_params(echidna, data):
    """
    Function to retrive the learned parameters for one pass
    """
    guide_trace = pyro.poutine.trace(echidna.guide).get_trace(*data)
    trained_model = pyro.poutine.replay(echidna.model, trace=guide_trace)
    trained_trace = pyro.poutine.trace(trained_model).get_trace(*data)
    params = trained_trace.nodes
    return params

def eta_posterior_estimates(echidna, data, num_samples=(int(1e3),)):
    """
    Posterior mean of eta
    """
    if isinstance(num_samples, tuple):
        num_samples = num_samples[0]
    X, _, pi, _ = data
    eta = torch.zeros([pi.shape[-1], X.shape[-1]])
    for _ in range(num_samples):
        params = get_learned_params(echidna, data)
        eta += F.softplus(params['eta']['value'].T)
    eta /= num_samples
    eta = eta.to(echidna.config.device)
    return eta

def c_posterior_estimates(eta, mt=True):
    """
    Posterior mean of c. Takes in posterior mean of eta
    """
    c = None
    if mt:
        c_shape = pyro.param('c_shape').squeeze(1)
        c = c_shape.unsqueeze(1) * eta.unsqueeze(0)
    else:
        c_shape = pyro.param('c_shape')
        c = c_shape * eta
    return c

def cov_posterior_estimate(inverse_gamma=False, num_samples=(int(1e3),)):
    """
    Posterior mean of covariance
    """
    corr_loc = pyro.param("corr_loc")
    corr_scale = pyro.param("corr_scale")
    corr_cov = torch.diag(corr_scale)
    corr_dist = dist.MultivariateNormal(corr_loc, corr_cov)
    transformed_dist = dist.TransformedDistribution(corr_dist, dist.transforms.CorrCholeskyTransform())
    chol_samples = transformed_dist.sample(num_samples)
    L_shape = pyro.param('scale_shape')
    L_rate = pyro.param('scale_rate')
    L = L_shape/L_rate
    
    scale = L[:, None] if not inverse_gamma else 1/L[:, None]
    cov = chol_samples.mean(0) * torch.sqrt(scale)
    cov = cov@cov.T
    
    return cov

def normalize_cov(cov):
    """Posterior mean of correlation matrix. Takes in estimated covariance."""
    std_dev = torch.sqrt(torch.diag(cov))
    outer_std_dev = torch.outer(std_dev, std_dev)
    corr_matrix = cov / outer_std_dev
    corr_matrix.fill_diagonal_(1)  # Set diagonal elements to 1
    return corr_matrix

def save_model(adata, model, overwrite=False, simulation=False):
    """
    Modified from Decipher with author permission:
    Achille Nazaret, https://github.com/azizilab/decipher/blob/main/decipher/tools/_decipher/data.py
    """
    create_echidna_uns_key(adata)
    
    run_id_key = "run_id_sim" if simulation else "run_id"
    run_id_key_hist = "run_id_sim_history" if simulation else "run_id_history"
    
    if run_id_key_hist not in adata.uns["echidna"]:
        adata.uns["echidna"][run_id_key_hist] = []
    else:
        run_hist = adata.uns["echidna"][run_id_key_hist]
        if isinstance(run_hist, np.ndarray):
            adata.uns["echidna"][run_id_key_hist] = list(run_hist)
    sim = "" if not simulation else "simulation "
    if run_id_key not in adata.uns["echidna"] or not overwrite:
        adata.uns["echidna"][run_id_key] = datetime.now().strftime("%Y%m%d-%H%M%S")
        adata.uns["echidna"][run_id_key_hist].append(adata.uns["echidna"][run_id_key])
        logging.info(f"Saving echidna {sim}model with run_id {adata.uns['echidna'][run_id_key]}.")
    else:
        logging.info(f"Overwriting existing echidna {sim}model.")

    model_run_id = adata.uns["echidna"][run_id_key]
    save_folder = ECHIDNA_GLOBALS["save_folder"]
    full_path = os.path.join(save_folder, model_run_id)
    os.makedirs(full_path, exist_ok=True)
    
    torch.save(model, os.path.join(full_path, "echidna_model.pt"))
    
    param_store = pyro.get_param_store()
    param_dict = {name: param_store[name].detach().cpu() for name in param_store.keys()}
    torch.save(param_dict, os.path.join(full_path, "echidna_model_param_store.pt"))
    
    adata.uns["echidna"]["config"] = model.config.to_dict()

def load_model(adata, save_folder=None, simulation=False):
    """
    Modified from Decipher with author permission:
    Achille Nazaret, https://github.com/azizilab/decipher/blob/main/decipher/tools/_decipher/data.py
    
    Load an echidna model whose name is stored in the given AnnData.

    `adata.uns["echidna"]["run_id"]` must be set to the name of the echidna model to load.

    Parameters
    ----------
    adata : sc.AnnData
        The annotated data matrix.
    
    save_folder: str
        The location of saved model parameters. Default to None.

    Returns
    -------
    model : Echidna
        The echidna model.
    """
    create_echidna_uns_key(adata)
    
    run_id_key = "run_id_sim" if simulation else "run_id"
    run_id_key_hist = "run_id_sim_history" if simulation else "run_id_history"
    
    if run_id_key not in adata.uns["echidna"]:
        sim = "" if not simulation else "simulation "
        raise ValueError(f"No echidna {sim}model has been saved for this AnnData object.")

    model_config = EchidnaConfig(**adata.uns["echidna"]["config"])
    model = Echidna(model_config)
    model_run_id = adata.uns["echidna"][run_id_key]
    save_folder_path = ECHIDNA_GLOBALS["save_folder"] if save_folder is None else save_folder
    full_path = os.path.join(save_folder_path, model_run_id) if save_folder is None else save_folder
    model = torch.load(os.path.join(full_path, "echidna_model.pt"), weights_only=False)
    
    pyro.clear_param_store()
    gc.collect()
    torch.cuda.empty_cache()
    param_store = pyro.get_param_store()
    param_dict = torch.load(os.path.join(full_path, "echidna_model_param_store.pt"), weights_only=False)
    for name, param in param_dict.items():
        if name in param_store:
            param_store[name] = param.to(model.config.device)
        else:
            pyro.param(name, param.to(model.config.device))

    torch.set_default_dtype(torch.float32)
    torch.set_default_device(model.config.device)

    return model


# echidna.tools.data.py

import pyro
import torch
import numpy as np
import pandas as pd
import scanpy as sc
import warnings, requests, json, re

from echidna.tools.utils import EchidnaConfig, _custom_sort
from echidna.utils import get_logger

logger = get_logger(__name__)

def pre_process(
    adata: sc.AnnData, 
    num_genes: int=None, 
    target_sum: float=None, 
    exclude_highly_expressed: bool=False, 
    n_comps: int=15, 
    phenograph_k: int=60, 
    n_neighbors: int=15,
    filepath: str=None,
    ) -> sc.AnnData:
    """Basic pre-processing pipeline. Choose parameters according to your data.
    
    Parameters
    ----------
    adata : sc.AnnData
        Unprocessed annotated data matrix.
    num_genes : int
        Number of highly expressed genes to use. Pass None if using
        all (recommended).
    target_sum : 
        Normalize to this total, defaults median library size.
    exclude_highly_expressed : bool
        Whether to exclude highly expressed genes.
    n_comps : int
        Number of principal components to use for PCA.
    phenograph_k : int
        Number of nearest neighbors to use in first step of graph
        construction.
    n_neighbors : int
        Number of nearest neighbors for UMAP.
    filepath : str
        If defined, will save the processed AnnData to the specified location.
        
    Returns
    -------
    adata : sc.AnnData
        Processed annotated data matrix.
    """
    from scipy.sparse import csr_matrix

    adata.X = adata.X.astype("float32")

    # highly variable genes
    if num_genes is not None:
        x_log = sc.pp.log1p(adata, copy=True, base=10)
        sc.pp.highly_variable_genes(x_log, n_top_genes=num_genes)
        adata = adata[:, x_log.var["highly_variable"]]

    if "counts" not in adata.layers:
        adata.layers["counts"] = adata.X.copy()
    sc.pp.calculate_qc_metrics(adata, inplace=True, layer="counts")

    # store the current "total_counts" under original_total_counts, 
    # which will not automatically be updated by scanpy in subsequent filtering steps
    adata.obs["original_total_counts"] = adata.obs["total_counts"].copy()

    # log10 original library size
    adata.obs["log10_original_total_counts"] = np.log10(adata.obs["original_total_counts"]).copy()

    # Normalize by median library size
    if target_sum is None:
        target_sum = np.median(adata.obs["original_total_counts"])
    sc.pp.normalize_total(adata, target_sum=target_sum, exclude_highly_expressed=exclude_highly_expressed)

    # log transform + 1 and updates adata.X
    sc.pp.log1p(adata)

    logger.info("Performing PCA...")
    sc.tl.pca(adata, n_comps=n_comps)
    logger.info("Calculating phenograph clusters...")
    sc.external.tl.phenograph(adata, clustering_algo="leiden", k=phenograph_k, seed=1)

    logger.info("Performing nearest neighbors search and calculating UMAP...")
    sc.pp.neighbors(adata, random_state=1, n_neighbors=n_neighbors)
    sc.tl.umap(adata, random_state=1)

    for sparse_mtx in adata.obsp:
        adata.obsp[sparse_mtx] = csr_matrix(adata.obsp[sparse_mtx])
    if filepath is not None:
        adata.write_h5ad(filepath)
    return adata

def filter_low_var_genes(
    adata: sc.AnnData,
    quantile: float=0.75,
    var_threshold: float=None,
    indices: bool=False,
) -> sc.AnnData:
    gene_variances = adata.X.var(axis=0)
    
    if var_threshold is not None:
        gene_filter = gene_variances > var_threshold # / (adata.X.mean(axis=0) + 1e-8)
    elif var_threshold is None:
        var_threshold = np.quantile(gene_variances, quantile)
        gene_filter = gene_variances > var_threshold
    if indices:
        return gene_filter
    return adata[:, gene_filter]

def train_val_split(adata, config):
    rng = np.random.default_rng(config.seed)
    
    tmp_idx, i = "index", 0
    while tmp_idx in adata.obs.columns:
        tmp_idx = f"index{i}"
        i+=1
    adata.obs.reset_index(names=tmp_idx, inplace=True)
    
    if config._is_multi:
        adata_vc = adata.obs[config.timepoint_label].value_counts()

        n_val = int(config.val_split * adata_vc.min())
        smallest_tp = adata_vc.index[adata_vc.argmin()]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            adata.obs["echidna_split"] = "train"
        for tp in adata_vc.index:
            tp_filter = adata.obs[config.timepoint_label] == tp
            cur_tp_index = adata.obs[tp_filter].index

            val_idx = rng.choice(cur_tp_index, n_val, replace=False)

            n_discard = len(cur_tp_index) - adata_vc.min()
            if n_discard > 0:
                train_idx = np.setdiff1d(cur_tp_index, val_idx)
                discard_idx = rng.choice(train_idx, n_discard, replace=False)
                adata.obs.loc[discard_idx, "echidna_split"] = "discard"

            adata.obs.loc[val_idx, "echidna_split"] = "validation"

        adata.obs["echidna_split"] = adata.obs["echidna_split"].astype("category")
    else:
        n_val = int(config.val_split * adata.shape[0])
        val_idx = rng.choice(adata.obs.index, n_val, replace=False)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            adata.obs["echidna_split"] = "train"
        adata.obs.loc[adata.obs.index[val_idx], "echidna_split"] = "validation"
        adata.obs["echidna_split"] = adata.obs["echidna_split"].astype("category")
        
    adata.obs.set_index(tmp_idx, inplace=True)
    
    logger.info(
        "Added `.obs['echidna_split']`: the Echidna train/validation split.\n"
        f" {n_val} cells in validation set."
    )

    return adata

def create_z_pi(adata, config):
    config = config.to_dict() if isinstance(config, EchidnaConfig) else config
    if config["clusters"] not in adata.obs.columns:
        raise ValueError(f"{config['clusters']} clustering not in AnnData obs")
    if not bool(config["_is_multi"]):
        z_obs_series = adata.obs[config["clusters"]].values
        pi_obs_series = np.unique(z_obs_series, return_counts=True)[1] / len(z_obs_series)
        z_obs = torch.from_numpy(np.array(z_obs_series)).to(torch.float32).to(config["device"])
        pi_obs = torch.from_numpy(pi_obs_series).to(torch.float32).to(config["device"])
    else:
        adata_tmp = adata[adata.obs["echidna_split"]!="discard"].copy()
        timepoints = np.unique(adata_tmp.obs[config["timepoint_label"]])
        if "timepoint_order" in adata_tmp.uns["echidna"]:
            timepoints = _custom_sort(timepoints, adata.uns["echidna"]["timepoint_order"])
        z = []
        pi = []
        for t in timepoints:
            z_tmp = adata_tmp.obs[adata_tmp.obs[config["timepoint_label"]]==t][config["clusters"]].values
            pi_tmp = torch.zeros(int(config["num_clusters"]), dtype=torch.float32)
            indices, counts = np.unique(z_tmp, return_counts=True)
            pi_proportions = counts / len(z_tmp)
            for i, p in zip(indices, pi_proportions): pi_tmp[i] = p
            z.append(torch.tensor(z_tmp, dtype=torch.int64))
            pi.append(pi_tmp)
        z_obs = torch.stack(z).to(torch.float32).to(config["device"])
        pi_obs = torch.stack(pi).to(torch.float32).to(config["device"])
    return pi_obs, z_obs

def match_genes(adata, Wdf):
    """Matches genes between AnnData and W.

    Parameters
    ----------
        adata : sc.AnnData
            Annotated data matrix.
        Wdf : pd.DataFrame
            DataFrame containing copy number counts, indexed by genes.
    """
    if Wdf.index.duplicated().any():
        raise ValueError("Duplicate indices in W. Make sure W is uniquely indexed by genes.")
    
    Wdf.dropna(inplace=True)
    matched_genes = adata.var.index.intersection(Wdf.index)
    adata.var["echidna_matched_genes"] = np.where(adata.var.index.isin(matched_genes), True, False)
    logger.info("Added `.var[echidna_matched_genes]` : Labled True for genes contained in W.")
    if len(Wdf.shape) > 1:
        Wdf.columns = [f"echidna_W_{c}" if "echidna_W_" not in c else c for c in Wdf.columns]
        col_name = Wdf.columns
    elif len(Wdf.shape) == 1:
        if Wdf.name is None:
            Wdf.name = "echidna_W_count"
        else:
            Wdf.name = "echidna_W_" + Wdf.name if "echidna_W_" not in Wdf.name else Wdf.name
        col_name = [Wdf.name]
    if len(np.intersect1d(adata.var.columns, col_name)) == 0:
        adata.var = adata.var.merge(Wdf, left_index=True, right_index=True, how="left")
        # display(merged_var)
        # adata = adata[:, ]
        for c in col_name:
            logger.info(f"Added `.var[{c}]` : CN entries for genes contained in W.")

def build_torch_tensors(adata, config):
    """
    Takes anndata and builds Torch tensors.
    """
    Wdf = adata.var[[c for c in adata.var.columns if "echidna_W_" in c]].dropna()
    if config._is_multi and "timepoint_order" in adata.uns["echidna"]:
        Wdf = _custom_sort(Wdf, adata.uns["echidna"]["timepoint_order"])
    W_obs = torch.from_numpy(Wdf.values).to(torch.float32).to(config.device)
    
    if W_obs.shape[-1] != config.num_timepoints:
        raise ValueError(
            "Number of W columns found in AnnData does not match"
            " number of timepoints, drop excess if needed."
            " Check columns in `.var` :", list(Wdf.columns)
        )
    if config._is_multi:
        W_obs = W_obs.T
        adata = adata[adata.obs["echidna_split"]!="discard"]
        tps = adata.obs[config.timepoint_label].unique()
        if "timepoint_order" in adata.uns["echidna"]:
            tps = _custom_sort(tps, adata.uns["echidna"]["timepoint_order"])
        x_list = [adata[tp == adata.obs[config.timepoint_label]].layers[config.counts_layer] for tp in tps]
        X_obs = torch.from_numpy(np.array(x_list)).to(torch.float32).to(config.device)
    elif not config._is_multi:
        W_obs = W_obs.flatten()
        X_obs = torch.from_numpy(adata.layers[config.counts_layer]).to(torch.float32).to(config.device)
    if config.clusters:
        pi_obs, z_obs = create_z_pi(adata, config)
        return X_obs, W_obs, pi_obs, z_obs
    return X_obs, W_obs

def fetch_genome():
    genes = pd.read_csv(
        "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/wgEncodeGencodeCompV46.txt.gz",
        delimiter="\t",
        header=None,
        names=["gene_id", "transcript_id", "chrom", "strand", "txStart", "txEnd", "cdsStart", "cdsEnd", "exonCount", "exonStarts", "exonEnds", "score", "geneName", "cdsStartStat", "cdsEndStat", "exonFrames"],
    )
    genes = genes[["chrom", "geneName", "txStart", "txEnd"]].drop_duplicates()
    genes_gb = genes.groupby(["chrom", "geneName"])
    min_tx = genes_gb.min()["txStart"].reset_index()
    max_tx = genes_gb.max()["txEnd"].reset_index()
    gene_coords = min_tx.merge(max_tx, on=["chrom", "geneName"])
    
    ensembl_to_gene_name = ensembl_conversion(list(gene_coords[gene_coords.geneName.str.startswith("ENS")].geneName))
    if ensembl_to_gene_name:
        def map_gene_names(row, mapping):
            new_name = mapping.get(row['geneName'], None)
            return new_name if new_name is not None else row["geneName"]

        gene_coords["geneName"] = gene_coords.apply(lambda row: map_gene_names(row, ensembl_to_gene_name), axis=1)
    
    return sort_chromosomes(gene_coords)

def ensembl_conversion(ensembl_genes):
    response = requests.post(
        "https://biotools.fr/mouse/ensembl_symbol_converter/",
        data={
            'api': 1,
            'ids': json.dumps(ensembl_genes)
        },
    )
    if response.status_code == 200:
        output = response.json()
    else:
        logger.error(f"Request failed with status code: {response.status_code}")
        return
    return output

def range_subset_pct(
    gene_range, chrom_region
):
    start, end = gene_range
    chrom_start, chrom_end = chrom_region
    overlap = max(0, min(end, chrom_end) - max(start, chrom_start))
    return overlap / (end - start) if (end - start) > 0 else 0

def range_subset(
    gene_range, chrom_region
):
    start, end = gene_range
    chrom_start, chrom_end = chrom_region
    return 1 if start >= chrom_start and end <= chrom_end else 0

def sort_chromosomes(df):
    def chrom_key(chrom):
        match = re.match(r"chr(\d+|X|Y)", chrom)
        if match:
            chrom = match.group(1)
            if chrom == "X":
                return (float("inf"), 1)  # X at the end but before Y
            elif chrom == "Y":
                return (float("inf"), 2)  # Y at the end
            else:
                return (int(chrom), 0)
        return (float("inf"), 3)  # Any unexpected value goes at the very end
    if isinstance(df, pd.DataFrame):
        df = df.copy()
        df["chrom_key"] = df["chrom"].apply(chrom_key)
        sort_cols = np.intersect1d(["bandStart", "txStart"], df.columns)
        df_sorted = df.sort_values(by=["chrom_key"] + list(sort_cols)).drop(columns="chrom_key")
        return df_sorted
    elif isinstance(df, pd.Series):
        sorted_index = sorted(df.index, key=chrom_key)
        return df[sorted_index]

def get_w(
    ichor: pd.DataFrame,
    genome: pd.DataFrame,
    timepoint: str,
    verbose: bool=False,
    chrom_label: str="chrom",
    gene_start: str="txStart",
    gene_end: str="txEnd",
    gene_name: str="geneName",
    weighted: bool=True,
):
    w = []
    genome = genome.drop_duplicates()
    ichor["chr"] = ichor["chr"].apply(lambda x: x if x.startswith("chr") else "chr" + x)
    
    func = range_subset_pct if weighted else range_subset
    
    seen = defaultdict(float)
    for _, cn_row in ichor.iterrows():

        chrom = genome[genome[chrom_label].astype(str) == cn_row["chr"]]
        chrom_region = (cn_row.start, cn_row.end)
        
        if verbose: print(f"------- Matching chromosome {cn_row.chr} and range {chrom_region}...")
        for _, gene in chrom.iterrows():
            gene_range = (gene[gene_start], gene[gene_end])
            pct = func(gene_range, chrom_region)
            
            if gene[gene_name] not in seen:
                seen[gene[gene_name]] = pct * cn_row["copy.number"]
            else:
                seen[gene[gene_name]] += pct * cn_row["copy.number"]

    W = pd.DataFrame({"geneName": list(seen.keys()), f"{timepoint}_count": list(seen.values())})
    return W